# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import logging
import pickle

import numpy as np
import pandas as pd

from pytest import approx, mark

from lenskit.als import ImplicitMFScorer
from lenskit.als._implicit import ImplicitMFConfig
from lenskit.data import Dataset, ItemList, RecQuery, from_interactions_df, load_movielens_df
from lenskit.pipeline import Pipeline, topn_pipeline
from lenskit.testing import BasicComponentTests, ScorerTests

_log = logging.getLogger(__name__)

simple_df = pd.DataFrame({"item": [1, 1, 2, 3], "user": [10, 12, 10, 13]})
simple_ds = from_interactions_df(simple_df)

simple_dfr = simple_df.assign(rating=[4.0, 3.0, 5.0, 2.0])
simple_dsr = from_interactions_df(simple_dfr)


class TestImplicitALS(BasicComponentTests, ScorerTests):
    component = ImplicitMFScorer
    expected_ndcg = 0.22

    def verify_models_equivalent(self, orig, copy):
        assert np.all(copy.user_embeddings == orig.user_embeddings)
        assert np.all(copy.item_embeddings == orig.item_embeddings)
        assert np.all(copy.items.index == orig.items.index)
        assert np.all(copy.users.index == orig.users.index)


def test_config_defaults():
    cfg = ImplicitMFConfig()
    assert cfg.embedding_size == 50


def test_config_es_alias():
    cfg = ImplicitMFConfig(features=72)  # type: ignore
    assert cfg.embedding_size == 72


def test_config_exp_ctor():
    cfg = ImplicitMFConfig(embedding_size_exp=5)  # type: ignore
    assert cfg.embedding_size == 32


def test_config_exp_dict():
    cfg = ImplicitMFConfig.model_validate({"embedding_size_exp": 10})
    assert cfg.embedding_size == 1024


def test_config_exp_dict_double():
    cfg = ImplicitMFConfig.model_validate({"embedding_size": 20, "embedding_size_exp": 10})
    assert cfg.embedding_size == 1024


def test_config_exp_json():
    cfg = ImplicitMFConfig.model_validate_json('{"embedding_size_exp": 2}')
    assert cfg.embedding_size == 4


def test_reconfigure_pipeline():
    algo = ImplicitMFScorer()
    pipe = topn_pipeline(algo)

    pipe_cfg = pipe.config

    cfg2 = pipe_cfg.merge_component_configs({"scorer": {"embedding_size_exp": 4}})
    assert cfg2.components["scorer"].config["embedding_size_exp"] == 4
    print(cfg2.components["scorer"].model_dump_json(indent=2))
    p2 = Pipeline.from_config(cfg2)
    scorer = p2.node("scorer")
    print(scorer.component.config.model_dump_json(indent=2))
    assert scorer.component.config.embedding_size == 16


def test_als_basic_build():
    algo = ImplicitMFScorer(features=20, epochs=10)
    algo.train(simple_ds)

    assert algo.users is not None
    assert algo.user_embeddings is not None

    assert set(algo.users.ids()) == set([10, 12, 13])
    assert set(algo.items.ids()) == set([1, 2, 3])
    assert algo.user_embeddings.shape == (3, 20)
    assert algo.item_embeddings.shape == (3, 20)


def test_als_predict_basic():
    algo = ImplicitMFScorer(features=20, epochs=10)
    algo.train(simple_ds)

    preds = algo(query=10, items=ItemList([3]))

    assert len(preds) == 1
    preds = preds.scores("pandas", index="ids")
    assert preds is not None
    assert preds.index[0] == 3
    assert preds.loc[3] >= -0.1
    assert preds.loc[3] <= 5


def test_als_predict_basic_for_new_ratings():
    """Test ImplicitMF ability to support new ratings"""
    algo = ImplicitMFScorer(features=20, epochs=10)
    algo.train(simple_ds)

    query = RecQuery(15, ItemList([1, 2]))
    preds = algo(query, ItemList([3]))

    assert len(preds) == 1
    preds = preds.scores("pandas", index="ids")
    assert preds is not None
    assert preds.index[0] == 3
    assert preds.loc[3] >= -0.1
    assert preds.loc[3] <= 5


def test_als_predict_basic_for_new_user_with_new_ratings():
    """
    Test if ImplicitMF predictions using the same ratings for a new user
    is the same as a user in the current simple_df dataset.
    """
    u = 10
    i = 3

    algo = ImplicitMFScorer(features=20, epochs=10, use_ratings=True)
    algo.train(simple_dsr)

    preds = algo(u, ItemList([i]))
    preds = preds.scores("pandas", index="ids")
    assert preds is not None

    query = RecQuery(1, ItemList([1, 2], rating=[4.0, 5.0]))
    new_preds = algo(query, ItemList([i]))
    new_preds = new_preds.scores("pandas", index="ids")
    assert new_preds is not None
    assert abs(preds.loc[i] - new_preds.loc[i]) <= 0.1


def test_als_predict_for_new_users_with_new_ratings(rng: np.random.Generator, ml_ds: Dataset):
    """
    Test if ImplicitMF predictions using the same ratings for a new user
    is the same as a user in ml-latest-small dataset.
    The test is run for more than one user.
    """
    n_users = 3
    n_items = 2
    new_u_id = -1

    np.random.seed(45)
    users = rng.choice(ml_ds.users.ids(), n_users)
    items = ItemList(rng.choice(ml_ds.items.ids(), n_items))

    algo = ImplicitMFScorer(features=20, epochs=10, use_ratings=False)
    algo.train(ml_ds)
    assert algo.users is not None
    assert algo.user_embeddings is not None

    _log.debug("Items: " + str(items))

    for u in users:
        _log.debug(f"user: {u}")
        preds = algo(u, items)
        preds = preds.scores("pandas", index="ids")
        assert preds is not None
        upos = algo.users.number(u)

        # get the user's rating series
        user_data = ml_ds.user_row(u)
        assert user_data is not None

        nr_info = user_data.to_df()
        ifs = algo.item_embeddings[user_data.numbers(vocabulary=algo.items), :]
        fit_uv = algo.user_embeddings[upos, :]
        nr_info["fit_recon"] = ifs @ fit_uv
        nr_info["fit_sqerr"] = np.square(algo.config.weight + 1.0 - nr_info["fit_recon"])

        _log.debug("user_features from fit:\n%s", fit_uv)
        new_uv, _new_off = algo.new_user_embedding(new_u_id, user_data)
        nr_info["new_recon"] = ifs @ new_uv
        nr_info["new_sqerr"] = np.square(algo.config.weight + 1.0 - nr_info["new_recon"])

        _log.debug("user features from new:\n%s", new_uv)

        _log.debug("training data reconstruction:\n%s", nr_info)

        query = RecQuery(-1, user_data)
        new_preds = algo(query, items)
        new_preds = new_preds.scores("pandas", index="ids")
        assert new_preds is not None

        _log.debug("preds: " + str(preds.values))
        _log.debug("new_preds: " + str(new_preds.values))
        _log.debug("------------")

        diffs = np.abs(preds - new_preds)
        assert all(diffs <= 0.1)


def test_als_recs_topn_for_new_users_with_new_ratings(
    rng, ml_ratings: pd.DataFrame, ml_ds: Dataset
):
    """
    Test if ImplicitMF topn recommendations using the same ratings for a new user
    is the same as a user in ml-latest-small dataset.
    The test is run for more than one user.
    """
    import scipy.stats as stats

    n_users = 10

    users = rng.choice(ml_ds.users.ids(), n_users).tolist()

    algo = ImplicitMFScorer(features=20, epochs=10, use_ratings=True)
    pipe = topn_pipeline(algo, n=10)
    pipe.train(ml_ds)
    assert algo.users is not None
    assert algo.user_embeddings is not None
    # _log.debug("Items: " + str(items))

    correlations = pd.Series(np.nan, index=users)
    for u in users:
        recs = pipe.run("recommender", query=u)
        assert isinstance(recs, ItemList)
        user_data = ml_ds.user_row(u)
        assert user_data is not None
        upos = algo.users.number(u)
        _log.info("user %s: %s ratings", u, len(user_data))

        _log.debug("user_features from fit: " + str(algo.user_embeddings[upos, :]))

        # get the user's rating series
        query = RecQuery(-1, user_data)

        new_recs = pipe.run("recommender", query=query)
        assert isinstance(new_recs, ItemList)

        # merge new & old recs
        old_scores = recs.scores("pandas", index="ids")
        assert old_scores is not None
        new_scores = new_recs.scores("pandas", index="ids")
        assert new_scores is not None

        old_scores, new_scores = old_scores.align(new_scores, join="outer", fill_value=-np.nan)

        tau = stats.kendalltau(old_scores, new_scores)
        _log.info("correlation for user %s: %f", u, tau.correlation)
        correlations.loc[u] = tau.correlation

    _log.debug("correlations: %s", correlations)

    assert not (any(correlations.isnull()))
    assert all(correlations >= 0.5)


def test_als_predict_bad_item():
    algo = ImplicitMFScorer(features=20, epochs=10)
    algo.train(simple_ds)

    preds = algo(10, ItemList([4]))
    assert len(preds) == 1
    preds = preds.scores("pandas", index="ids")
    assert preds is not None
    assert preds.index[0] == 4
    assert np.isnan(preds.loc[4])


def test_als_predict_bad_user():
    algo = ImplicitMFScorer(features=20, epochs=10)
    algo.train(simple_ds)

    preds = algo(50, ItemList([3]))
    assert len(preds) == 1
    preds = preds.scores("pandas", index="ids")
    assert preds is not None
    assert preds.index[0] == 3
    assert np.isnan(preds.loc[3])


def test_als_predict_no_user_features_basic(ml_ratings: pd.DataFrame, ml_ds: Dataset):
    np.random.seed(45)
    u = np.random.choice(ml_ds.users.ids(), 1)[0]
    items = np.random.choice(ml_ds.items.ids(), 2)

    algo = ImplicitMFScorer(features=5, epochs=10)
    algo.train(ml_ds)
    preds = algo(u, ItemList(items))
    preds = preds.scores("pandas", index="ids")
    assert preds is not None

    user_data = ml_ds.user_row(u)

    algo_no_user_features = ImplicitMFScorer(features=5, epochs=10, user_embeddings=False)
    algo_no_user_features.train(ml_ds)
    query = RecQuery(u, user_data)
    preds_no_user_features = algo_no_user_features(query, ItemList(items))
    preds_no_user_features = preds_no_user_features.scores("pandas", index="ids")
    assert preds_no_user_features is not None

    assert algo_no_user_features.user_embeddings is None
    assert preds_no_user_features.values == approx(preds, abs=0.1)
    diffs = np.abs(preds - preds_no_user_features)
    assert all(diffs <= 0.1)


def test_als_train_large(ml_ds: Dataset):
    algo = ImplicitMFScorer(features=20, epochs=20, use_ratings=False)
    algo.train(ml_ds)

    assert algo.users is not None
    assert algo.user_embeddings is not None
    assert len(algo.users.index) == ml_ds.user_count
    assert len(algo.items.index) == ml_ds.item_count
    assert algo.user_embeddings.shape == (ml_ds.user_count, 20)
    assert algo.item_embeddings.shape == (ml_ds.item_count, 20)


def test_als_train_large_noratings(ml_ds: Dataset):
    algo = ImplicitMFScorer(features=20, epochs=20)
    algo.train(ml_ds)

    assert algo.users is not None
    assert algo.user_embeddings is not None
    assert len(algo.users.index) == ml_ds.user_count
    assert len(algo.items.index) == ml_ds.item_count
    assert algo.user_embeddings.shape == (ml_ds.user_count, 20)
    assert algo.item_embeddings.shape == (ml_ds.item_count, 20)


def test_als_train_large_ratings(ml_ds):
    algo = ImplicitMFScorer(features=20, epochs=20, use_ratings=True)
    algo.train(ml_ds)

    assert algo.users is not None
    assert algo.user_embeddings is not None
    assert len(algo.users.index) == ml_ds.user_count
    assert len(algo.items.index) == ml_ds.item_count
    assert algo.user_embeddings.shape == (ml_ds.user_count, 20)
    assert algo.item_embeddings.shape == (ml_ds.item_count, 20)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Hypothesis strategies that generate arrays and lists.
"""

from typing import Literal

import numpy as np
import scipy.sparse as sps

import hypothesis.extra.numpy as nph
import hypothesis.strategies as st
from hypothesis import assume

from lenskit.data import ItemList
from lenskit.math.sparse import torch_sparse_from_scipy


@st.composite
def coo_arrays(
    draw,
    shape=None,
    dtype=nph.floating_dtypes(endianness="=", sizes=[32, 64]),
    elements=st.floats(-1e6, 1e6, allow_nan=False, allow_infinity=False, width=32),
) -> sps.coo_array:
    if shape is None:
        shape = st.tuples(st.integers(1, 100), st.integers(1, 100))

    if isinstance(shape, st.SearchStrategy):
        shape = draw(shape)

    if not isinstance(shape, tuple):
        shape = shape, shape
    rows, cols = shape
    if isinstance(rows, st.SearchStrategy):
        rows = draw(rows)
    if isinstance(cols, st.SearchStrategy):
        cols = draw(cols)

    mask = draw(nph.arrays(np.bool_, (rows, cols)))
    # at least one nonzero value
    assume(np.any(mask))
    nnz = int(np.sum(mask))

    ris, cis = np.nonzero(mask)

    vs = draw(
        nph.arrays(dtype, nnz, elements=elements),
    )

    return sps.coo_array((vs, (ris, cis)), shape=(rows, cols))


@st.composite
def sparse_arrays(draw, *, layout="csr", **kwargs):
    if isinstance(layout, list):
        layout = st.sampled_from(layout)
    if isinstance(layout, st.SearchStrategy):
        layout = draw(layout)

    M: sps.coo_array = draw(coo_arrays(**kwargs))

    match layout:
        case "csr":
            return M.tocsr()
        case "csc":
            return M.tocsc()
        case "coo":
            return M
        case _:
            raise ValueError(f"invalid layout {layout}")


@st.composite
def sparse_tensors(draw, *, layout="csr", **kwargs):
    if isinstance(layout, list):
        layout = st.sampled_from(layout)
    if isinstance(layout, st.SearchStrategy):
        layout = draw(layout)

    M: sps.coo_array = draw(coo_arrays(**kwargs))
    return torch_sparse_from_scipy(M, layout)  # type: ignore


@st.composite
def scored_lists(
    draw: st.DrawFn,
    *,
    n: int | tuple[int, int] | st.SearchStrategy[int] = st.integers(0, 1000),
    scores: st.SearchStrategy[float] | Literal["gaussian"] | None = None,
) -> ItemList:
    """
    Hypothesis generator that produces scored lists.
    """
    if isinstance(n, st.SearchStrategy):
        n = draw(n)
    elif isinstance(n, tuple):
        n = draw(st.integers(*n))

    ids = np.arange(1, n + 1, dtype=np.int32)
    if scores == "gaussian":
        seed = draw(st.integers(0))
        rng = np.random.default_rng(seed)
        xs = np.exp(rng.normal(size=n))
    else:
        xs = draw(
            nph.arrays(nph.floating_dtypes(endianness="=", sizes=[32, 64]), n, elements=scores)
        )
    return ItemList(item_ids=ids, scores=xs)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
k-NN recommender models.
"""

from .item import ItemKNNConfig, ItemKNNScorer
from .user import UserKNNConfig, UserKNNScorer

__all__ = ["ItemKNNScorer", "ItemKNNConfig", "UserKNNScorer", "UserKNNConfig"]



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import numpy as np
import pandas as pd

from pytest import approx, mark

from lenskit.data import ItemList
from lenskit.metrics.ranking import RecipRank
from lenskit.testing import demo_recs  # noqa: F401


def _test_rr(items, rel, **kw):
    recs = ItemList(items, ordered=True)
    truth = ItemList(rel)
    return RecipRank(**kw).measure_list(recs, truth)


def test_mrr_empty_zero():
    rr = _test_rr([], [1, 3])
    assert rr == approx(0)


def test_mrr_norel_zero():
    "no relevant items -> zero"
    rr = _test_rr([1, 2, 3], [4, 5])
    assert rr == approx(0)


def test_mrr_first_one():
    "first relevant -> one"
    rr = _test_rr([1, 2, 3], [1, 4])
    assert rr == approx(1.0)


def test_mrr_second_one_half():
    "second relevant -> 0.5"
    rr = _test_rr([1, 2, 3], [5, 2, 3])
    assert rr == approx(0.5)


def test_mrr_series():
    "second relevant -> 0.5 in pd series"
    rr = _test_rr(pd.Series([1, 2, 3]), pd.Series([5, 2, 3]))
    assert rr == approx(0.5)


def test_mrr_series_idx():
    "second relevant -> 0.5 in pd series w/ index"
    rr = _test_rr(pd.Series([1, 2, 3]), pd.Index([5, 2, 3]))
    assert rr == approx(0.5)


def test_mrr_array_late():
    "deep -> 0.1"
    rr = _test_rr(np.arange(1, 21, 1, "u4"), [20, 10])
    assert rr == approx(0.1)


def test_mrr_k_trunc():
    rr = _test_rr(np.arange(1, 21, 1, "u4"), [20, 10], k=5)
    assert rr == approx(0.0)

    rr = _test_rr(np.arange(1, 21, 1, "u4"), [20, 10, 5], k=5)
    assert rr == approx(0.2)


def test_mrr_k_short():
    rr = _test_rr(np.arange(1, 5, 1, "u4"), [2], k=10)
    assert rr == approx(0.5)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Pipeline caching support.
"""

from __future__ import annotations

from collections.abc import Callable, Mapping
from dataclasses import dataclass, field
from typing import Any

from pydantic import JsonValue, TypeAdapter

from lenskit.logging import get_logger

from .components import ComponentConstructor
from .types import type_string

_log = get_logger(__name__)


@dataclass(frozen=True)
class PipelineCacheKey:
    ctor_name: str
    ctor: Callable[..., Any] = field(hash=False)

    # this is not hashed because it's a dict, we can just do the slow ==.
    # there aren't that many of them.
    config: Mapping[str, JsonValue] | None = field(hash=False, default=None)


class PipelineCache:
    """
    A cache to share components between pipelines.

    This cache can be used by :class:`~lenskit.pipeline.PipelineBuilder` to
    cache pipeline component instances between multiple pipelines using the same
    components with the same configuration.
    """

    _cache: dict[PipelineCacheKey, Any]

    def __init__(self):
        self._cache = {}

    def _make_key(self, ctor: ComponentConstructor[Any], config: object | None):
        name = type_string(ctor)  # type: ignore
        if config is not None:
            config = TypeAdapter[Any](ctor.config_class()).dump_python(config, mode="json")
            assert isinstance(config, dict)

        return PipelineCacheKey(name, ctor, config)

    def get_cached(self, ctor: ComponentConstructor[Any], config: object | None):
        key = self._make_key(ctor, object)
        return self._cache.get(key, None)

    def cache(self, ctor: ComponentConstructor[Any], config: object | None, instance: object):
        key = self._make_key(ctor, object)
        self._cache[key] = instance

    def get_instance(self, ctor: ComponentConstructor[Any], config: object | None):
        """
        Get a component instance from the cache, creating if it necessry.
        """
        key = self._make_key(ctor, config)
        instance = self._cache.get(key, None)
        if instance is None:
            instance = ctor(config)
            self._cache[key] = instance
            _log.debug(
                "instantiated component", component=ctor, config=config, n_cached=len(self._cache)
            )
        else:
            _log.debug(
                "found cached component", component=ctor, config=config, n_cached=len(self._cache)
            )
        return instance



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import numpy as np
import pandas as pd
import torch

import hypothesis.extra.numpy as nph
import hypothesis.strategies as st
from hypothesis import given
from pytest import mark

from lenskit.data import ItemList
from lenskit.testing import scored_lists


def test_topn_empty():
    il = ItemList(scores=0)

    top = il.top_n()
    assert top.ordered
    assert len(top) == 0


@given(scored_lists())
def test_top_all(items):
    top = items.top_n()
    assert len(top) == len(items) - np.sum(np.isnan(items.scores()))
    assert top.ordered

    scores = top.scores()
    assert scores is not None
    diffs = np.diff(scores)
    assert np.all((diffs <= 0) | np.isnan(diffs))


@given(scored_lists(), st.integers(min_value=1))
def test_top_n(items, n):
    top = items.top_n(n)
    assert len(top) == min(n, len(items) - np.sum(np.isnan(items.scores())))
    assert top.ordered

    scores = top.scores()
    assert scores is not None
    diffs = np.diff(scores)
    assert np.all((diffs <= 0) | np.isnan(diffs))


@given(scored_lists(), st.integers(min_value=1))
def test_top_n_field(items, n):
    items = ItemList(items, scores=False, rating=items.scores())
    top = items.top_n(n, scores="rating")
    assert len(top) == min(n, len(items) - np.sum(np.isnan(items.field("rating"))))
    assert top.ordered

    rates = top.field("rating")
    assert rates is not None
    diffs = np.diff(rates)
    assert np.all((diffs <= 0) | np.isnan(diffs))


@given(scored_lists(), st.integers(min_value=1))
def test_top_n_keys(items, n):
    keys = np.random.randn(len(items))
    top = items.top_n(n, scores=keys)
    assert len(top) == min(n, len(items))
    assert top.ordered

    # the items should correspond to keys in decreasing order
    keys = pd.Series(keys, index=items.ids())
    keys = keys.reindex(top.ids())
    diffs = np.diff(keys)
    assert np.all((diffs <= 0) | np.isnan(diffs))


@mark.skipif(not torch.cuda.is_available(), reason="CUDA test")
@given(scored_lists())
def test_top_all_cuda(items):
    items = ItemList(items, scores=items.scores(format="torch").to("cuda"))
    top = items.top_n()
    assert len(top) == len(items) - np.sum(np.isnan(items.scores()))
    assert top.ordered

    scores = top.scores()
    assert scores is not None
    diffs = np.diff(scores)
    assert np.all((diffs <= 0) | np.isnan(diffs))


@mark.skipif(not torch.cuda.is_available(), reason="CUDA test")
@given(scored_lists(), st.integers(min_value=1))
def test_top_n_cuda(items, n):
    items = ItemList(items, scores=items.scores(format="torch").to("cuda"))
    top = items.top_n(n)
    assert len(top) == min(n, len(items) - np.sum(np.isnan(items.scores())))
    assert top.ordered

    scores = top.scores()
    assert scores is not None
    diffs = np.diff(scores)
    assert np.all((diffs <= 0) | np.isnan(diffs))



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import numpy as np
import pyarrow as pa
import pyarrow.compute as pc

import hypothesis.strategies as st
from hypothesis import given

from lenskit._accel import data


@given(st.lists(st.integers(0, 1000)), st.integers(0, 5000))
def test_neg_mask(indices, n):
    iarr = pa.array(indices, pa.int32())
    imax = pc.max(iarr).as_py()
    if imax is not None and imax >= n:
        n = imax + 1

    mask = data.negative_mask(n, iarr)
    np_mask = mask.to_numpy(zero_copy_only=False)

    assert len(mask) == n
    assert np.sum(np_mask) == n - len(set(indices))
    assert np.all(~np_mask[indices])



from lenskit.data.matrix import SparseRowArray
from lenskit.data.types import NPMatrix
from lenskit.logging import Progress

def train_explicit_matrix(
    matrix: SparseRowArray,
    this: NPMatrix,
    other: NPMatrix,
    reg: float,
    pb: Progress | None,
) -> float: ...
def train_implicit_matrix(
    matrix: SparseRowArray,
    this: NPMatrix,
    other: NPMatrix,
    otor: NPMatrix,
    pb: Progress | None,
) -> float: ...



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

# pyright: basic
import numpy as np
import pandas as pd
import pyarrow as pa
import torch
from scipy.sparse import csr_array

from pytest import approx, mark, raises, skip, warns

from lenskit.data import DatasetBuilder, from_interactions_df
from lenskit.data.matrix import SparseRowArray
from lenskit.data.schema import AttrLayout
from lenskit.diagnostics import DataError, DataWarning
from lenskit.testing import ml_test_dir

FRUITS = ["apple", "banana", "orange", "lemon", "mango"]


def test_item_scalar_series():
    dsb = DatasetBuilder()

    items = pd.read_csv(ml_test_dir / "movies.csv")
    items = items.rename(columns={"movieId": "item_id"}).set_index("item_id")

    dsb.add_entities("item", items.index.values)
    dsb.add_scalar_attribute("item", "title", items["title"])
    sa = dsb.schema.entities["item"].attributes["title"]
    assert sa.layout == AttrLayout.SCALAR

    ds = dsb.build()

    df = ds.entities("item").pandas().set_index("item_id")
    assert "title" in df.columns
    ds_ts, orig_ts = df["title"].align(items["title"])
    assert np.all(ds_ts == orig_ts)

    assert ds.entities("item").attribute("title").is_scalar
    df = ds.entities("item").attribute("title").pandas()
    assert isinstance(df, pd.Series)
    assert np.all(df == items["title"])


def test_item_scalar_df():
    dsb = DatasetBuilder()

    items = pd.read_csv(ml_test_dir / "movies.csv")
    items = items.rename(columns={"movieId": "item_id"})

    dsb.add_entities("item", items["item_id"].values)
    dsb.add_scalar_attribute("item", "title", items[["item_id", "title"]])
    sa = dsb.schema.entities["item"].attributes["title"]
    assert sa.layout == AttrLayout.SCALAR

    ds = dsb.build()

    df = ds.entities("item").pandas().set_index("item_id")
    assert "title" in df.columns
    ds_ts, orig_ts = df["title"].align(items.set_index("item_id")["title"])
    assert np.all(ds_ts == orig_ts)

    assert ds.entities("item").attribute("title").is_scalar
    df = ds.entities("item").attribute("title").pandas()
    assert isinstance(df, pd.Series)
    ds_ts, orig_ts = df.align(items.set_index("item_id")["title"])
    assert np.all(ds_ts == orig_ts)


def test_item_scalar_array():
    dsb = DatasetBuilder()

    items = pd.read_csv(ml_test_dir / "movies.csv")
    items = items.rename(columns={"movieId": "item_id"})

    dsb.add_entities("item", items["item_id"].values)
    dsb.add_scalar_attribute("item", "title", items["item_id"], items["title"])
    sa = dsb.schema.entities["item"].attributes["title"]
    assert sa.layout == AttrLayout.SCALAR

    ds = dsb.build()

    df = ds.entities("item").pandas().set_index("item_id")
    assert "title" in df.columns
    ds_ts, orig_ts = df["title"].align(items.set_index("item_id")["title"])
    assert np.all(ds_ts == orig_ts)

    assert ds.entities("item").attribute("title").is_scalar
    df = ds.entities("item").attribute("title").pandas()
    assert isinstance(df, pd.Series)
    ds_ts, orig_ts = df.align(items.set_index("item_id")["title"])
    assert np.all(ds_ts == orig_ts)


def test_item_scalar_series_arrays():
    dsb = DatasetBuilder()

    items = pd.read_csv(ml_test_dir / "movies.csv")
    items = items.rename(columns={"movieId": "item_id"})

    dsb.add_entities("item", items["item_id"].values)
    dsb.add_scalar_attribute("item", "title", items["item_id"], items["title"])
    sa = dsb.schema.entities["item"].attributes["title"]
    assert sa.layout == AttrLayout.SCALAR

    ds = dsb.build()

    df = ds.entities("item").pandas().set_index("item_id")
    assert "title" in df.columns
    ds_ts, orig_ts = df["title"].align(items.set_index("item_id")["title"])
    assert np.all(ds_ts == orig_ts)

    assert ds.entities("item").attribute("title").is_scalar
    df = ds.entities("item").attribute("title").pandas()
    assert isinstance(df, pd.Series)
    ds_ts, orig_ts = df.align(items.set_index("item_id")["title"])
    assert np.all(ds_ts == orig_ts)


def test_scalar_attribute_alignment():
    interactions = pd.DataFrame(
        {
            "user_id": [10, 10, 11, 12],
            "item_id": [1, 3, 2, 6],
        }
    )

    # attributes come in order [1, 6, 3], not matching entity order
    protected_df = pd.DataFrame(
        {
            "item_id": [1, 6, 3],
            "protected": [1, 0, 1],
        }
    )

    ds = from_interactions_df(interactions)
    builder = DatasetBuilder(ds)
    builder.add_scalar_attribute("item", "protected", protected_df)
    ds = builder.build()

    attributes_df = ds.entities("item").attribute("protected").pandas()

    # check alignment: values match original protected_df
    expected = protected_df.set_index("item_id")["protected"]
    assert attributes_df.equals(expected.reindex(attributes_df.index))

    # test alignment in numpy
    attributes_np = ds.entities("item").attribute("protected").numpy()
    entity_ids = ds.entities("item").vocabulary.ids(np.arange(len(attributes_np)))
    attributes_df = expected.reindex(entity_ids, fill_value=None)

    assert np.all(np.isclose(attributes_df.values, attributes_np, equal_nan=True))


@mark.xfail(reason="attributes at insert time not yet implemented")
def test_item_insert_scalar_df():
    dsb = DatasetBuilder()

    items = pd.read_csv(ml_test_dir / "movies.csv")
    items = items.rename(columns={"movieId": "item_id"})

    dsb.add_entities("item", items[["item_id", "title"]])
    sa = dsb.schema.entities["item"].attributes["title"]
    assert sa.layout == AttrLayout.SCALAR

    ds = dsb.build()

    df = ds.entities("item").pandas().set_index("item_id")
    assert "title" in df.columns
    ds_ts, orig_ts = df["title"].align(items.set_index("item_id")["title"])
    assert np.all(ds_ts == orig_ts)

    assert ds.entities("item").attribute("title").is_scalar
    df = ds.entities("item").attribute("title").pandas()
    assert isinstance(df, pd.Series)
    ds_ts, orig_ts = df.align(items.set_index("item_id")["title"])
    assert np.all(ds_ts == orig_ts)


@mark.xfail(reason="attributes at insert time not yet implemented")
def test_item_update_titles():
    dsb = DatasetBuilder()

    items = pd.read_csv(ml_test_dir / "movies.csv")
    items = items.rename(columns={"movieId": "item_id"})

    dsb.add_entities("item", items)

    dsb.add_scalar_attribute(
        "item", "title", pd.Series({2: "Board Game Adventure", 110: "Briarheart"})
    )

    ds = dsb.build()
    df = ds.entities("item").attribute("title").pandas()
    assert df.loc[1, "title"] == "Toy Story (1995)"
    assert df.loc[2, "title"] == "Board Game Adventure"
    assert df.loc[110, "title"] == "Briarheart"


def test_item_list_series():
    dsb = DatasetBuilder()

    items = pd.read_csv(ml_test_dir / "movies.csv")
    items = items.rename(columns={"movieId": "item_id"}).set_index("item_id")

    genres = items["genres"].str.split("|")
    g_counts = genres.apply(len)

    dsb.add_entities("item", items.index.values)
    dsb.add_list_attribute("item", "genres", genres)
    va = dsb.schema.entities["item"].attributes["genres"]
    assert va.layout == AttrLayout.LIST

    ds = dsb.build()

    assert ds.entities("item").attribute("genres").is_list
    arr = ds.entities("item").attribute("genres").arrow()
    if isinstance(arr, pa.ChunkedArray):
        arr = arr.combine_chunks()
    assert isinstance(arr, pa.ListArray)
    assert len(arr) == len(genres)
    assert np.all(
        arr.value_lengths().fill_null(0).to_numpy()
        == g_counts.reindex(ds.entities("item").ids()).values
    )

    gs = ds.entities("item").attribute("genres").pandas()
    assert np.all(gs.index == ds.items.ids())
    gs, gs2 = gs.align(genres, join="outer")
    assert np.all(gs.notnull())
    assert np.all(gs2.notnull())
    gs = gs.apply(lambda gl: ",".join(sorted(gl)))
    gs2 = gs2.apply(lambda gl: ",".join(sorted(gl)))
    assert np.all(gs == gs2)


def test_item_list_random():
    dsb = DatasetBuilder()

    items = pd.read_csv(ml_test_dir / "movies.csv")
    items = items.rename(columns={"movieId": "item_id"}).set_index("item_id")

    genres = items["genres"].str.split("|")

    genres = genres.sample(n=200)

    dsb.add_entities("item", items.index.values)
    dsb.add_list_attribute("item", "genres", genres.index.values, genres.tolist())

    va = dsb.schema.entities["item"].attributes["genres"]
    assert va.layout == AttrLayout.LIST

    ds = dsb.build()

    assert ds.entities("item").attribute("genres").is_list

    gs = ds.entities("item").attribute("genres").pandas()

    gcomp, gscomp = genres.align(gs, "left")
    assert np.all(gcomp.apply(len) == gscomp.apply(len))


def test_item_list_df():
    dsb = DatasetBuilder()

    items = pd.read_csv(ml_test_dir / "movies.csv")
    items = items.rename(columns={"movieId": "item_id"})
    items["genres"] = items["genres"].str.split("|")
    g_counts = items.set_index("item_id")["genres"].apply(len)

    dsb.add_entities("item", items["item_id"].values)
    dsb.add_list_attribute("item", "genres", items[["item_id", "genres"]])
    va = dsb.schema.entities["item"].attributes["genres"]
    assert va.layout == AttrLayout.LIST

    ds = dsb.build()
    assert ds.entities("item").attribute("genres").is_list
    arr = ds.entities("item").attribute("genres").arrow()
    if isinstance(arr, pa.ChunkedArray):
        arr = arr.combine_chunks()
    assert isinstance(arr, pa.ListArray)
    assert len(arr) == len(items)
    assert np.all(
        arr.value_lengths().fill_null(0).to_numpy()
        == g_counts.reindex(ds.entities("item").ids()).values
    )


@mark.xfail(reason="attributes at insert time not yet implemented")
def test_item_initial_list_df():
    dsb = DatasetBuilder()

    items = pd.read_csv(ml_test_dir / "movies.csv")
    items = items.rename(columns={"movieId": "item_id"})
    items["genres"] = items["genres"].str.split("|")

    dsb.add_entities("item", items)
    va = dsb.schema.entities["item"].attributes["genres"]
    assert va.layout == AttrLayout.LIST

    ds = dsb.build()

    assert ds.entities("item").attribute("genres").is_list

    arr = ds.entities("item").attribute("title").arrow()
    assert isinstance(arr, pa.StringArray)
    assert len(arr)

    arr = ds.entities("item").attribute("genres").arrow()
    assert isinstance(arr, pa.ListArray)
    assert len(arr) == len(items)


def test_item_vector(rng: np.random.Generator, ml_ratings: pd.DataFrame):
    dsb = DatasetBuilder()

    item_ids = ml_ratings["item_id"].unique()
    dsb.add_entities("item", item_ids)

    items = rng.choice(item_ids, 500, replace=False)
    vec = rng.standard_normal((500, 20))
    dsb.add_vector_attribute("item", "embedding", items, vec)
    va = dsb.schema.entities["item"].attributes["embedding"]
    assert va.layout == AttrLayout.VECTOR
    assert va.vector_size == 20

    tbl = dsb._tables["item"]
    assert tbl is not None
    col = tbl.column("embedding")
    assert len(col) == tbl.num_rows
    valid = col.is_valid().to_numpy(zero_copy_only=False)
    assert np.sum(valid) == 500
    col = col.combine_chunks()
    lens = col.value_lengths().fill_null(0).to_numpy()
    assert np.all(lens[valid] == 20)

    ds = dsb.build()
    assert ds.entities("item").attribute("embedding").is_vector
    assert ds.entities("item").attribute("embedding").dim_names is None

    arr = ds.entities("item").attribute("embedding").arrow()
    if isinstance(arr, pa.ChunkedArray):
        arr = arr.combine_chunks()
    assert isinstance(arr, pa.FixedSizeListArray)
    assert np.sum(np.asarray(arr.is_valid())) == 500
    assert np.all(np.asarray(arr.value_lengths().drop_null()) == 20)

    arr = ds.entities("item").attribute("embedding").numpy()
    assert isinstance(arr, np.ndarray)
    assert arr.shape == (len(item_ids), 20)
    assert np.sum(np.isfinite(arr)) == 500 * 20

    lim = ds.entities("item").attribute("embedding").drop_null()
    idx = pd.Index(lim.ids())
    v2 = lim.numpy()
    for i, id in enumerate(items):
        i2 = idx.get_loc(id)
        print(id, i, i2)
        assert np.all(vec[i, :] == v2[i2, :])

    arr = ds.entities("item").attribute("embedding").pandas(missing="omit")
    assert isinstance(arr, pd.DataFrame)
    assert np.all(arr.columns == np.arange(20))
    assert arr.shape == (500, 20)
    assert set(arr.index) == set(items)


def test_item_vector_names(rng: np.random.Generator, ml_ratings: pd.DataFrame):
    dsb = DatasetBuilder()

    item_ids = ml_ratings["item_id"].unique()
    dsb.add_entities("item", item_ids)

    items = rng.choice(item_ids, 500, replace=False)
    vec = rng.standard_normal((500, 5))
    dsb.add_vector_attribute("item", "embedding", items, vec, dim_names=FRUITS)
    va = dsb.schema.entities["item"].attributes["embedding"]
    assert va.layout == AttrLayout.VECTOR

    ds = dsb.build()
    assert ds.entities("item").attribute("embedding").is_vector
    assert ds.entities("item").attribute("embedding").dim_names == FRUITS

    arr = ds.entities("item").attribute("embedding").arrow()
    if isinstance(arr, pa.ChunkedArray):
        arr = arr.combine_chunks()
    assert isinstance(arr, pa.FixedSizeListArray)
    assert np.sum(np.asarray(arr.is_valid())) == 500
    assert np.all(np.asarray(arr.value_lengths().drop_null()) == 5)

    arr = ds.entities("item").attribute("embedding").numpy()
    assert isinstance(arr, np.ndarray)
    assert arr.shape == (len(item_ids), 5)
    assert np.sum(np.isfinite(arr)) == 500 * 5

    arr = ds.entities("item").attribute("embedding").pandas(missing="omit")
    assert isinstance(arr, pd.DataFrame)
    assert np.all(arr.columns == FRUITS)
    assert arr.shape == (500, 5)
    assert set(arr.index) == set(items)


def test_item_vector_attr_subset(rng: np.random.Generator, ml_ratings: pd.DataFrame):
    dsb = DatasetBuilder()

    item_ids = ml_ratings["item_id"].unique()
    dsb.add_entities("item", item_ids)

    items = rng.choice(item_ids, 500, replace=False)
    vec = rng.standard_normal((500, 5))
    dsb.add_vector_attribute("item", "embedding", items, vec, dim_names=FRUITS)
    va = dsb.schema.entities["item"].attributes["embedding"]
    assert va.layout == AttrLayout.VECTOR

    ds = dsb.build()

    query_items = set(rng.choice(items, 100, replace=False))
    query_items |= set(rng.choice(item_ids, 50, replace=False))
    query_items = np.array(list(query_items))
    q_known = np.isin(query_items, items)

    subset = ds.entities("item").select(ids=query_items)
    assert len(subset) == len(query_items)

    ss_attrs = subset.attribute("embedding")
    assert ss_attrs.is_vector
    assert ss_attrs.dim_names == FRUITS

    assert np.all(ss_attrs.ids() == query_items)
    assert np.all(ss_attrs.numbers() == ds.items.numbers(query_items))

    arr = ss_attrs.arrow()
    if isinstance(arr, pa.ChunkedArray):
        arr = arr.combine_chunks()
    assert isinstance(arr, pa.FixedSizeListArray)
    assert len(arr) == len(query_items)
    assert np.all(arr.is_valid().to_numpy(zero_copy_only=False) == q_known)

    df = ss_attrs.pandas(missing="omit")
    assert isinstance(df, pd.DataFrame)
    assert np.all(df.columns == FRUITS)
    assert df.shape == (np.sum(q_known), 5)


def test_add_invalid_attribute_name():
    dsb = DatasetBuilder()

    items = pd.read_csv(ml_test_dir / "movies.csv")
    items = items.rename(columns={"movieId": "item_id"}).set_index("item_id")

    dsb.add_entities("item", items.index.values)

    invalid_attributes = ["_title", "title_id"]
    for attribute_name in invalid_attributes:
        with raises(ValueError, match="invalid"):
            dsb.add_scalar_attribute("item", attribute_name, items["title"])


def test_item_sparse_attribute(rng: np.random.Generator, ml_ratings: pd.DataFrame):
    dsb = DatasetBuilder()

    movies = pd.read_csv(ml_test_dir / "movies.csv")
    movies = movies.rename(columns={"movieId": "item_id"}).set_index("item_id")

    dsb.add_entities("item", movies.index)

    genre_lists = movies["genres"].str.split("|")
    genres = genre_lists.reset_index().explode("genres", ignore_index=True)
    gindex = pd.Index(np.unique(genres["genres"]))

    ids = movies.index.values.copy()
    rng.shuffle(ids)
    idx2 = pd.Index(ids)
    ig_rows = idx2.get_indexer_for(genres["item_id"])
    ig_cols = gindex.get_indexer_for(genres["genres"])
    ig_vals = np.ones(len(ig_rows), np.int32)

    arr = csr_array((ig_vals, (ig_rows, ig_cols)))
    dsb.add_vector_attribute("item", "genres", idx2, arr, dim_names=gindex)

    ga = dsb.schema.entities["item"].attributes["genres"]
    assert ga.layout == AttrLayout.SPARSE

    ds = dsb.build()

    assert ds.entities("item").attribute("genres").is_sparse
    assert ds.entities("item").attribute("genres").dim_names == gindex.values.tolist()

    mat = ds.entities("item").attribute("genres").scipy()
    assert isinstance(mat, csr_array)
    assert mat.nnz == arr.nnz

    m2 = ds.entities("item").select(ids=idx2.values).attribute("genres").scipy()
    assert isinstance(mat, csr_array)
    assert np.all(m2.indptr == arr.indptr)

    tensor = ds.entities("item").attribute("genres").torch()
    assert isinstance(tensor, torch.Tensor)
    assert tensor.is_sparse_csr
    assert len(tensor.values()) == arr.nnz

    arr = ds.entities("item").attribute("genres").arrow()
    assert isinstance(arr, SparseRowArray)

    for movie in rng.choice(movies.index.values, 50, replace=False):
        row = ds.items.number(movie)
        start = mat.indptr[row]
        end = mat.indptr[row + 1]
        cols = mat.indices[start:end]
        mgs = gindex[cols].tolist()
        assert mgs == genre_lists.loc[movie]



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import logging
import pickle

import numpy as np
import pandas as pd
from numpy.typing import NDArray
from scipy.spatial.distance import jaccard
from scipy.special import softmax
from scipy.stats import kendalltau, permutation_test

import hypothesis.extra.numpy as nph
import hypothesis.strategies as st
from hypothesis import given, settings
from pytest import mark

from lenskit import batch
from lenskit.als import ImplicitMFScorer
from lenskit.basic import PopScorer
from lenskit.basic.topn import TopNRanker
from lenskit.data import Dataset, ItemList
from lenskit.logging import get_logger
from lenskit.operations import recommend
from lenskit.pipeline import topn_pipeline
from lenskit.splitting import simple_test_pair
from lenskit.stochastic import StochasticTopNRanker
from lenskit.testing import BasicComponentTests, ScorerTests, scored_lists
from lenskit.training import TrainingOptions

_log = get_logger(__name__)


class TestSoftmax(BasicComponentTests):
    component = StochasticTopNRanker


@mark.filterwarnings("error:divide by zero")
@given(scored_lists(), st.sampled_from(["linear", "softmax"]))
def test_unlimited_ranking(items: ItemList, transform):
    topn = StochasticTopNRanker(transform=transform)
    ranked = topn(items=items)

    ids = items.ids()
    scores = items.scores("numpy")
    assert scores is not None
    invalid = ~np.isfinite(scores)

    try:
        assert isinstance(ranked, ItemList)
        assert len(ranked) <= len(items)
        assert ranked.ordered
        # all valid items are included
        assert len(ranked) == np.sum(~invalid)

        # the set of valid items matches
        assert set(ids[~invalid]) == set(ranked.ids())

        # the scores match
        rank_s = ranked.scores("pandas", index="ids")
        assert rank_s is not None
        src_s = items.scores("pandas", index="ids")
        assert src_s is not None

        # make sure the scores were preserved properly
        rank_s, src_s = rank_s.align(src_s, "left")
        assert not np.any(np.isnan(src_s))
        assert np.all(rank_s == src_s)
    except AssertionError as e:
        e.add_note("ranked {} items ({} invalid)".format(len(ids), np.sum(invalid)))
        raise e


@given(st.integers(min_value=1, max_value=100), scored_lists())
def test_configured_truncation(n, items: ItemList):
    topn = StochasticTopNRanker(n=n)
    ranked = topn(items=items)

    ids = items.ids()
    scores = items.scores("numpy")
    assert scores is not None
    invalid = ~np.isfinite(scores)
    _log.info("top %d of %d items, %d invalid", n, len(ids), np.sum(invalid))

    val_items = items[~invalid]

    assert isinstance(ranked, ItemList)
    assert ranked.ordered
    assert len(ranked) == min(n, len(val_items))

    # the scores match
    rank_s = ranked.scores("pandas", index="ids")
    assert rank_s is not None
    src_s = items.scores("pandas", index="ids")
    assert src_s is not None
    src_s = src_s[src_s.notna()]

    # make sure the scores were preserved properly
    rank_s, src_s = rank_s.align(src_s, "left")
    assert not np.any(np.isnan(src_s))
    assert np.all(rank_s == src_s)


@given(st.integers(min_value=1, max_value=100), scored_lists())
def test_runtime_truncation(n, items: ItemList):
    topn = StochasticTopNRanker(rng="user")
    ranked = topn(items=items, n=n)

    ids = items.ids()
    scores = items.scores("numpy")
    assert scores is not None
    invalid = ~np.isfinite(scores)
    _log.info("top %d of %d items, %d invalid", n, len(ids), np.sum(invalid))

    val_items = items[~invalid]

    assert isinstance(ranked, ItemList)
    assert ranked.ordered
    assert len(ranked) == min(n, len(val_items))

    # the scores match
    rank_s = ranked.scores("pandas", index="ids")
    assert rank_s is not None
    src_s = items.scores("pandas", index="ids")
    assert src_s is not None
    src_s = src_s[src_s.notna()]

    # make sure the scores were preserved properly
    rank_s, src_s = rank_s.align(src_s, "left")
    assert not np.any(np.isnan(src_s))
    assert np.all(rank_s == src_s)


@given(
    scored_lists(
        n=st.integers(100, 5000), scores=st.floats(-5, 15, width=32, allow_infinity=False)
    ),
    st.floats(0.1, 10000),
)
def test_overflow(items: ItemList, scale: float):
    topn = StochasticTopNRanker(transform="softmax", scale=scale)
    ranked = topn(items=items, include_weights=True)

    ids = items.ids()
    scores = items.scores("numpy")
    assert scores is not None
    assert np.all(np.isfinite(scores))

    sm = softmax(scores)
    assert np.all(np.isfinite(sm))
    assert np.all(sm >= 0)
    assert np.all(sm <= 1)

    k2 = topn._compute_keys(scores, topn._rng_factory(None))
    assert np.all(np.isfinite(k2))

    try:
        assert isinstance(ranked, ItemList)
        assert len(ranked) == len(items)
        assert ranked.ordered

        weights = ranked.field("weight")
        assert weights is not None
        assert np.all(np.isfinite(weights))

        # the scores match
        rank_s = ranked.scores("pandas", index="ids")
        assert rank_s is not None
        src_s = items.scores("pandas", index="ids")
        assert src_s is not None

        # make sure the scores were preserved properly
        rank_s, src_s = rank_s.align(src_s, "left")
        assert not np.any(np.isnan(src_s))
        assert np.all(rank_s == src_s)
    except AssertionError as e:
        e.add_note("ranked {} items".format(len(ids)))
        raise e


def test_stochasticity(rng):
    "Test that softmax is varying but order-consistent"
    iids = np.arange(500)
    scores = rng.normal(size=500)
    scores = np.square(scores)
    items = ItemList(item_ids=iids, scores=scores)
    size = 50

    TRIALS = 100
    topn = StochasticTopNRanker(n=size)

    _log.info("testing stochastic ranking: top %d of %d", size, len(items))

    ranks = np.full((size, TRIALS), -1, dtype=np.int64)
    scores = np.full((size, TRIALS), np.nan, dtype=np.float64)
    for i in range(TRIALS):
        ranked = topn(items)
        assert len(ranked) == size
        ranks[:, i] = ranked.ids()
        scores[:, i] = ranked.scores()

    id_counts = np.array([len(np.unique(ranks[i, :])) for i in range(size)])
    try:
        # at least half the positions should have more than 5 different items show up
        assert np.sum(id_counts < 5) <= size / 2
    except AssertionError as e:
        _log.info("failed test with n=%d, N=%d", size, len(items))
        _log.info("item counts: %s", id_counts)
        _log.info("items:\n%s", items.to_df())
        raise e

    # We want to test that it is usually putting things in the correct order.
    # We'll do this by computing Kendall's tau between the sampled ranking and
    # the scores of those items.  We want most of the lists (90%) to have
    # correlations statistically significnatly greater than zero (items tend to
    # be in the correct order).
    pvals = []
    taus = []
    for i in range(TRIALS):
        r_items = ranks[:, i]
        ranked = r_items >= 0
        r_items = r_items[ranked]

        rii = items[r_items]
        trs = size - np.arange(size)

        tau = kendalltau(trs, rii.scores(), alternative="greater")
        taus.append(tau.statistic)
        pvals.append(tau.pvalue)
        _log.info("trial %d: 𝜏=%.3f, p=%.3f", i, tau.statistic, tau.pvalue)

    pvals = np.array(pvals)
    _log.info("trial p-value statistics: mean=%.3f, median=%.3f", np.mean(pvals), np.median(pvals))
    # do 90% of trials pass the test?
    assert np.mean(pvals < 0.05) >= 0.9


@mark.flaky(retries=3)
def test_scale_affects_ranking(ml_ds: Dataset):
    """
    Test that different softmax scales produce different levels of ranking variation.
    """
    rng = np.random.default_rng()
    pipe = topn_pipeline(ImplicitMFScorer(embedding_size=32, weight=5))
    pipe.train(ml_ds, TrainingOptions(rng=rng))

    seed1, seed2, seed3 = rng.bit_generator.random_raw(3)

    topn = TopNRanker()
    samp_frac = StochasticTopNRanker(scale=0.01, rng=int(seed1))
    samp_one = StochasticTopNRanker(scale=1, rng=int(seed2))
    samp_hundred = StochasticTopNRanker(scale=50, rng=int(seed3))

    jc_frac = []
    jc_one = []
    jc_hundred = []

    for uid in rng.choice(ml_ds.users.ids(), size=500):
        ilist = recommend(pipe, uid)
        rl_topn = topn(items=ilist, n=20)
        rl_frac = samp_frac(items=ilist, n=20)
        assert len(rl_frac) == 20
        rl_one = samp_one(items=ilist, n=20)
        assert len(rl_one) == 20
        rl_hundred = samp_hundred(items=ilist, n=20)
        assert len(rl_hundred) == 20

        jc_frac.append(_jaccard(rl_topn, rl_frac))
        jc_one.append(_jaccard(rl_topn, rl_one))
        jc_hundred.append(_jaccard(rl_topn, rl_hundred))

    # high-temp should agree less than flat
    assert np.mean(jc_frac) < np.mean(jc_one)
    # low-temp should agree more than flat
    assert np.mean(jc_hundred) > np.mean(jc_one)


def _jaccard(il1: ItemList, il2: ItemList) -> float:
    s1 = set(il1.ids())
    s2 = set(il2.ids())

    num = len(s1 & s2)
    denom = len(s1 | s2)

    return num / denom



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from __future__ import annotations

from typing import Any

import numpy as np
import pyarrow as pa
from scipy.sparse import coo_array, csr_array

from hypothesis import given
from pytest import mark, raises

from lenskit._accel import sparse_row_debug_type, sparse_structure_debug_large
from lenskit.data import Dataset
from lenskit.data.matrix import (
    SparseIndexListType,
    SparseIndexType,
    SparseRowArray,
    SparseRowType,
)
from lenskit.logging import get_logger
from lenskit.testing import sparse_arrays

_log = get_logger(__name__)


@given(sparse_arrays())
def test_sparse_from_csr(csr: csr_array[Any, tuple[int, int]]):
    arr = SparseRowArray.from_scipy(csr)
    assert isinstance(arr, SparseRowArray)
    assert isinstance(arr.type, SparseRowType)

    assert pa.types.is_list(arr.storage.type)
    assert len(arr) == csr.shape[0]
    assert arr.dimension == csr.shape[1]
    assert len(arr.offsets) == csr.shape[0] + 1
    assert len(arr.indices) == csr.nnz
    assert len(arr.values) == csr.nnz
    assert arr.offsets.to_numpy()[0] == 0
    assert arr.offsets.to_numpy()[-1] == csr.nnz
    assert np.all(arr.offsets.to_numpy() == csr.indptr)
    assert isinstance(arr.indices.type, SparseIndexType)


@given(sparse_arrays())
def test_sparse_row_indices(csr: csr_array[Any, tuple[int, int]]):
    arr = SparseRowArray.from_scipy(csr)
    assert isinstance(arr, SparseRowArray)
    assert isinstance(arr.type, SparseRowType)

    for i in range(csr.shape[0]):
        indices = arr.row_indices(i)
        row = csr[[i], :]
        assert len(indices) == row.nnz
        assert np.all(indices == row.indices)


@given(sparse_arrays())
def test_csr_to_rust(csr: csr_array[Any, tuple[int, int]]):
    "Test that we can send CSRs to Rust"
    arr = SparseRowArray.from_scipy(csr)
    assert isinstance(arr, SparseRowArray)
    assert isinstance(arr.type, SparseRowType)

    dt, nr, nc = sparse_row_debug_type(arr)
    _log.info("returned data type: %s", dt, nr=nr, nc=nc)
    assert (nr, nc) == csr.shape


@given(sparse_arrays())
def test_csr_structure(csr: csr_array[Any, tuple[int, int]]):
    "Test that we can extract the structure from a CSR array."
    arr = SparseRowArray.from_scipy(csr)
    assert arr.has_values

    struct = arr.structure()
    assert not struct.has_values
    assert struct.shape == arr.shape
    assert np.all(struct.indices.to_numpy() == arr.indices.to_numpy())


@given(sparse_arrays())
def test_csr_structure_rust_large(csr: csr_array[Any, tuple[int, int]]):
    "Test that Rust can decode and convert large CSR structures."
    arr = SparseRowArray.from_scipy(csr).structure()
    assert isinstance(arr, SparseRowArray)
    assert isinstance(arr.type, SparseIndexListType)

    nr, nc, nnz = sparse_structure_debug_large(arr)
    assert (nr, nc) == csr.shape
    assert nnz == csr.nnz


def test_csr_structure_rust_ml(ml_ds: Dataset):
    "Test that Rust can decode and convert MovieLens CSR"
    arr = ml_ds.interactions().matrix().csr_structure(format="arrow")
    assert isinstance(arr, SparseRowArray)
    assert isinstance(arr.type, SparseIndexListType)

    nr, nc, nnz = sparse_structure_debug_large(arr)
    assert nr == ml_ds.user_count
    assert nc == ml_ds.item_count
    assert nnz == ml_ds.interaction_count


@given(sparse_arrays())
def test_sparse_to_csr(csr: csr_array[Any, tuple[int, int]]):
    arr = SparseRowArray.from_scipy(csr)

    csr2 = arr.to_scipy()
    assert csr2.shape == csr.shape
    assert csr2.nnz == csr.nnz
    assert np.all(csr2.indptr == csr.indptr)
    assert np.all(csr2.indices == csr.indices)
    assert np.all(csr2.data == csr.data.astype("f4"))


@given(sparse_arrays())
def test_sparse_structure(csr: csr_array[Any, tuple[int, int]]):
    nr, nc = csr.shape

    arr = SparseRowArray.from_arrays(csr.indptr, csr.indices, shape=(nr, nc))
    assert len(arr) == nr
    assert arr.dimension == nc
    assert np.all(arr.offsets.to_numpy() == csr.indptr)
    assert arr.values is None


@given(sparse_arrays())
def test_sparse_to_coo(csr: csr_array[Any, tuple[int, int]]):
    arr = SparseRowArray.from_scipy(csr)

    tbl = arr.to_coo()
    coo = csr.tocoo()
    assert np.all(tbl.column("row").to_numpy() == coo.row)
    assert np.all(tbl.column("col").to_numpy() == coo.col)
    assert np.all(tbl.column("value").to_numpy() == coo.data)


@given(sparse_arrays())
def test_sparse_to_coo_struct(csr: csr_array[Any, tuple[int, int]]):
    arr = SparseRowArray.from_scipy(csr).structure()

    tbl = arr.to_coo()
    coo = csr.tocoo()
    assert np.all(tbl.column("row").to_numpy() == coo.row)
    assert np.all(tbl.column("col").to_numpy() == coo.col)
    assert "value" not in tbl.column_names


@given(sparse_arrays())
def test_sparse_structure_rust(csr: csr_array[Any, tuple[int, int]]):
    nr, nc = csr.shape

    arr = SparseRowArray.from_arrays(csr.indptr, csr.indices, shape=(nr, nc))

    dt, nr, nc = sparse_row_debug_type(arr)
    _log.info("returned data type: %s", dt, nr=nr, nc=nc)
    assert (nr, nc) == csr.shape


@given(sparse_arrays())
def test_sparse_from_legacy(csr: csr_array[Any, tuple[int, int]]):
    tbl = pa.ListArray.from_arrays(
        pa.array(csr.indptr.astype("i4")),
        pa.StructArray.from_arrays(
            [pa.array(csr.indices.astype("i4")), pa.array(csr.data)], ["index", "value"]
        ),
    )  # type: ignore
    nr, nc = csr.shape

    arr = SparseRowArray.from_array(tbl, nc)
    assert isinstance(arr, SparseRowArray)
    assert len(arr) == nr
    assert arr.dimension == nc


@given(sparse_arrays())
def test_sparse_from_legacy_i8(csr: csr_array[Any, tuple[int, int]]):
    tbl = pa.ListArray.from_arrays(
        pa.array(csr.indptr.astype("i4")),
        pa.StructArray.from_arrays(
            [pa.array(csr.indices.astype("i8")), pa.array(csr.data)], ["index", "value"]
        ),
    )  # type: ignore
    nr, nc = csr.shape

    arr = SparseRowArray.from_array(tbl, nc)
    assert isinstance(arr, SparseRowArray)
    assert len(arr) == nr
    assert arr.dimension == nc
    assert pa.types.is_int32(arr.storage.values.field(0).storage.type)


def test_sparse_type_direct_mismatch():
    t = SparseRowType(20)

    with raises(ValueError, match="mismatch"):
        SparseRowType.from_type(t, 30)


def test_sparse_type_extract_mismatch():
    t = SparseRowType(20)
    print("type", t)
    print("storage type", t.storage_type)

    with raises(ValueError, match="mismatch"):
        SparseRowType.from_type(t.storage_type, 30)


def test_sparse_type_extract_no_dimension():
    t = SparseRowType(20)
    print("type", t)
    print("storage type", t.storage_type)

    t2 = SparseRowType.from_type(t.storage_type)
    assert t2.dimension == t.dimension


def test_sparse_type_extract_missing_dimension():
    orig_t = pa.list_(
        pa.struct(
            [
                ("index", pa.int32()),
                ("value", pa.float32()),
            ]
        )
    )

    with raises(TypeError):
        SparseRowType.from_type(orig_t)


def test_sparse_type_bad_index():
    orig_t = pa.list_(
        pa.struct(
            [
                ("index", pa.int64()),
                ("value", pa.float32()),
            ]
        )
    )

    with raises(TypeError):
        SparseRowType.from_type(orig_t)


def test_sparse_type_not_struct():
    orig_t = pa.list_(pa.float32())

    with raises(TypeError):
        SparseRowType.from_type(orig_t)


def test_sparse_type_not_list():
    orig_t = pa.struct(
        [
            ("index", pa.int64()),
            ("value", pa.float32()),
        ]
    )

    with raises(TypeError):
        SparseRowType.from_type(orig_t)


@given(sparse_arrays())
def test_sparse_transpose(csr: csr_array[Any, tuple[int, int]]):
    arr = SparseRowArray.from_scipy(csr)
    anr, anc = arr.shape
    assert (anr, anc) == csr.shape

    arrT = arr.transpose()
    assert isinstance(arrT, SparseRowArray)
    atnr, atnc = arrT.shape
    assert atnr == anc
    assert atnc == anr

    mt = arrT.to_scipy()
    assert np.all(mt.todense() == csr.T.todense())


@given(sparse_arrays())
def test_sparse_transpose_struct(csr: csr_array[Any, tuple[int, int]]):
    arr = SparseRowArray.from_scipy(csr).structure()
    anr, anc = arr.shape
    assert (anr, anc) == csr.shape

    arrT = arr.transpose()
    assert isinstance(arrT, SparseRowArray)
    atnr, atnc = arrT.shape
    assert atnr == anc
    assert atnc == anr
    assert not arrT.has_values

    assert np.all(arrT.indices.to_numpy() < atnc)

    coo_trans = csr.T.tocsr()
    # check that row sizes are correct
    assert np.all(arrT.offsets.to_numpy() == coo_trans.indptr)
    # check that columns appear correct # of times
    sp_cc = np.zeros(coo_trans.shape[1], dtype=np.int32)
    np.add.at(sp_cc, coo_trans.indices, 1)
    ar_cc = np.zeros(coo_trans.shape[1], dtype=np.int32)
    np.add.at(ar_cc, arrT.indices, 1)
    assert np.all(ar_cc == sp_cc)


def test_transpose_ratings(ml_ds: Dataset):
    rates = ml_ds.interactions().matrix().csr_structure(format="arrow")

    tr = rates.transpose()
    assert not tr.has_values
    assert tr.shape == (ml_ds.item_count, ml_ds.user_count)
    assert np.all(np.diff(tr.offsets) == ml_ds.item_stats()["rating_count"].values)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import nox


@nox.session(venv_backend="uv")
@nox.parametrize("torch", ["2.4", "2.5", "2.6", "2.7", "2.8"])
def test(session, torch):
    session.install(f"torch ~={torch}.0", "-e", ".", "--group", "test")
    opts = session.posargs
    if not opts:
        opts = ["-m", "not slow", "tests"]
    session.run("pytest", *opts)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Load Amazon ratings data from Julian McAuley's group.
"""

import re
from collections.abc import Generator
from io import BufferedReader
from os import PathLike
from pathlib import Path
from typing import BinaryIO

import pyarrow as pa
import pyarrow.compute as pc
from pyarrow.csv import ConvertOptions, ReadOptions, open_csv
from xopen import xopen

from lenskit.logging import get_logger

from .builder import DatasetBuilder
from .dataset import Dataset

_log = get_logger(__name__)


def load_amazon_ratings(*files: Path | str | PathLike[str]) -> Dataset:
    """
    Load an Amazon ratings CSV file into a dataset.  Files may be compressed by
    any compressor supported by :mod:`xopen` and PyArrow.

    The era of data (2014, 2018, or 2023) is auto-detected from file content.

    Args:
        files:
            The source rating files. Each one will be appeded in turn to the
            loader.
    """

    dsb = DatasetBuilder()
    dsb.add_entity_class("user")
    dsb.add_relationship_class("rating", ["user", "item"], allow_repeats=False, interaction=True)

    version = "AZ?"
    category = "Unknown"

    for file in files:
        file = Path(file)
        log = _log.bind(file=str(file))
        log.debug("opening file")

        with xopen(file, "rb", threads=1) as xf, BufferedReader(xf) as bf:  # type: ignore
            block = bf.peek(4096)
            log.debug("file header: %r", block[:20])
            if block.startswith(b"user_id,parent_asin,"):
                reader = open_az_2023(bf)
                version = "AZ23"
            elif block[:1] == b"A":
                reader = open_az_2014(bf)
                version = "AZ14"
            else:
                reader = open_az_2018(bf)
                version = "AZ18"

            m = re.match(r"^(?:ratings_)?(.*?)\.", file.name)
            if m is not None:
                category = m[1]

            # update the name appropriately
            name = f"{version}-{category}"
            if dsb.name is None:
                dsb.schema.name = name
            elif dsb.name != name:
                dsb.schema.name = f"{version}-Mixed"

            log.info("reading %s %s", version, category)

            for tbl in reader:
                dsb.add_interactions("rating", tbl, missing="insert")

    return dsb.build()


def open_az_2023(input: BinaryIO) -> Generator[pa.Table]:
    # 2023 Amazon data: has labels, we will ignore history
    with open_csv(
        input,
        convert_options=ConvertOptions(
            include_columns=["user_id", "parent_asin", "rating", "timestamp"],
            column_types={"rating": pa.float32()},
        ),
    ) as reader:
        for batch in reader:
            batch = batch.rename_columns(["user_id", "item_id", "rating", "timestamp"])
            columns = {c: batch.column(c) for c in batch.column_names}
            columns["timestamp"] = pc.cast(batch.column("timestamp"), pa.timestamp("ms"))
            yield pa.Table.from_pydict(columns)


def open_az_2014(input: BinaryIO) -> Generator[pa.Table]:
    # 2014 Amazon data: user, item, rating, timestamp
    with open_csv(
        input,
        read_options=ReadOptions(column_names=["user_id", "item_id", "rating", "timestamp"]),
        convert_options=ConvertOptions(
            column_types={"rating": pa.float32()},
        ),
    ) as reader:
        for batch in reader:
            columns = {c: batch.column(c) for c in batch.column_names}
            columns["timestamp"] = pc.cast(batch.column("timestamp"), pa.timestamp("s"))
            yield pa.Table.from_pydict(columns)


def open_az_2018(input: BinaryIO) -> Generator[pa.Table]:
    # 2018 Amazon data: user, item, rating, timestamp
    with open_csv(
        input,
        read_options=ReadOptions(
            block_size=16 * 1024 * 1024,
            column_names=["item_id", "user_id", "rating", "timestamp"],
        ),
        convert_options=ConvertOptions(
            column_types={"rating": pa.float32()},
        ),
    ) as reader:
        for batch in reader:
            columns = {c: batch.column(c) for c in batch.column_names}
            columns["timestamp"] = pc.cast(batch.column("timestamp"), pa.timestamp("s"))
            yield pa.Table.from_pydict(columns)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from pathlib import Path

from pydantic import BaseModel

from pytest import mark

from lenskit.config import load_config_data

TEST_DIR = Path(__file__).parent


class HelloModel(BaseModel):
    subject: str


@mark.parametrize("ext", ["json", "toml", "yml", "yaml"])
def test_load_raw(ext):
    file = TEST_DIR / f"hello.{ext}"

    data = load_config_data(file)
    assert isinstance(data, dict)
    assert data["subject"] == "world"


@mark.parametrize("ext", ["json", "toml", "yml", "yaml"])
def test_load_model(ext):
    file = TEST_DIR / f"hello.{ext}"

    data = load_config_data(file, HelloModel)
    assert isinstance(data, HelloModel)
    assert data.subject == "world"



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from pathlib import Path

import numpy as np
import pyarrow as pa

from pytest import fixture, mark, skip

from lenskit.data import Dataset

ML_32M = Path("data/ml-32m")


@fixture(scope="module")
def ml_32m():
    if not ML_32M.exists():
        skip("ML32M not available")

    yield Dataset.load(ML_32M)


@mark.benchmark()
def test_neg_unverified(rng: np.random.Generator, ml_ds: Dataset, benchmark):
    matrix = ml_ds.interactions().matrix()

    users = rng.choice(ml_ds.user_count, 8192, replace=True)
    users = np.require(users, "i4")

    def sample():
        _items = matrix.sample_negatives(users, verify=False, rng=rng)

    benchmark(sample)


@mark.benchmark()
def test_neg_verified(rng: np.random.Generator, ml_ds: Dataset, benchmark):
    matrix = ml_ds.interactions().matrix()

    users = rng.choice(ml_ds.user_count, 8192, replace=True)
    users = np.require(users, "i4")

    def sample():
        _items = matrix.sample_negatives(users, rng=rng)

    benchmark(sample)


@mark.benchmark()
def test_neg_vweighted(rng: np.random.Generator, ml_ds: Dataset, benchmark):
    matrix = ml_ds.interactions().matrix()

    users = rng.choice(ml_ds.user_count, 8192, replace=True)
    users = np.require(users, "i4")

    def sample():
        _items = matrix.sample_negatives(users, weighting="popularity", rng=rng)

    benchmark(sample)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import logging
import pickle
from pathlib import Path

import numpy as np
import pandas as pd
import torch
from numpy.typing import NDArray
from scipy import linalg as la

import pytest
from pytest import approx, fixture, mark

from lenskit.basic.history import UserTrainingHistoryLookup
from lenskit.data import ItemList, ItemListCollection, UserIDKey, Vocabulary, from_interactions_df
from lenskit.diagnostics import DataWarning
from lenskit.knn.item import ItemKNNScorer
from lenskit.pipeline import topn_pipeline
from lenskit.pipeline.common import predict_pipeline
from lenskit.testing import BasicComponentTests, ScorerTests

_log = logging.getLogger(__name__)

simple_ratings = pd.DataFrame.from_records(
    [
        (1, 6, 4.0),
        (2, 6, 2.0),
        (1, 7, 3.0),
        (2, 7, 2.0),
        (3, 7, 5.0),
        (4, 7, 2.0),
        (1, 8, 3.0),
        (2, 8, 4.0),
        (3, 8, 3.0),
        (4, 8, 2.0),
        (5, 8, 3.0),
        (6, 8, 2.0),
        (1, 9, 3.0),
        (3, 9, 4.0),
    ],
    columns=["user_id", "item_id", "rating"],
)
simple_ds = from_interactions_df(simple_ratings)


@fixture
def ml_subset(ml_ratings):
    "Fixture that returns a subset of the MovieLens database."
    icounts = ml_ratings.groupby("item_id").rating.count()
    top = icounts.nlargest(500)
    top_rates = ml_ratings[ml_ratings["item_id"].isin(top.index)]
    _log.info("top 500 items yield %d of %d ratings", len(top_rates), len(ml_ratings))
    return top_rates


class TestItemKNN(BasicComponentTests, ScorerTests):
    can_score = "some"
    component = ItemKNNScorer
    expected_rmse = (0.85, 0.95)
    expected_ndcg = 0.03

    def verify_models_equivalent(self, orig: ItemKNNScorer, copy: ItemKNNScorer):
        o_mat = orig.sim_matrix.to_scipy()
        r_mat = copy.sim_matrix.to_scipy()
        assert all(np.logical_not(np.isnan(r_mat.data)))
        assert all(r_mat.data > 0)
        # a little tolerance
        assert all(r_mat.data < 1 + 1.0e-6)

        assert copy.items == orig.items

        if orig.item_means is not None:
            assert copy.item_means is not None
            assert all(copy.item_means == orig.item_means)

        assert all(copy.item_counts == orig.item_counts)
        assert copy.item_counts.sum() == len(r_mat.data)
        assert len(r_mat.data) == len(r_mat.data)
        assert all(r_mat.indptr == o_mat.indptr)
        assert r_mat.data == approx(o_mat.data)

        assert all(r_mat.indptr == o_mat.indptr)


def test_ii_config():
    model = ItemKNNScorer(k=30)
    cfg = model.dump_config()
    print(cfg)
    assert cfg["feedback"] == "explicit"
    assert cfg["max_nbrs"] == 30


def test_ii_train():
    algo = ItemKNNScorer(k=30, save_nbrs=500)
    algo.train(simple_ds)

    assert isinstance(algo.item_means, np.ndarray)
    assert isinstance(algo.item_counts, np.ndarray)
    matrix = algo.sim_matrix.to_scipy()

    test_means = simple_ratings.groupby("item_id")["rating"].mean()
    test_means = test_means.reindex(algo.items.ids())
    assert np.all(algo.item_means == test_means.values.astype("f4"))

    # 6 is a neighbor of 7
    six, seven = algo.items.numbers([6, 7])
    _log.info("six: %d", six)
    _log.info("seven: %d", seven)
    _log.info("matrix: %s", algo.sim_matrix)
    assert matrix[six, seven] > 0
    # and has the correct score
    six_v = simple_ratings[simple_ratings["item_id"] == 6].set_index("user_id").rating
    six_v = six_v - six_v.mean()
    seven_v = simple_ratings[simple_ratings["item_id"] == 7].set_index("user_id").rating
    seven_v = seven_v - seven_v.mean()
    denom = la.norm(six_v.values) * la.norm(seven_v.values)
    six_v, seven_v = six_v.align(seven_v, join="inner")
    num = six_v.dot(seven_v)
    assert matrix[six, seven] == approx(num / denom, 0.01)  # type: ignore

    assert all(np.logical_not(np.isnan(matrix.data)))
    assert all(matrix.data > 0)
    # a little tolerance
    assert all(matrix.data < 1 + 1.0e-6)


def test_ii_train_unbounded():
    algo = ItemKNNScorer(k=30)
    algo.train(simple_ds)

    matrix = algo.sim_matrix.to_scipy()
    assert all(np.logical_not(np.isnan(matrix.data)))
    assert all(matrix.data > 0)
    # a little tolerance
    assert all(matrix.data < 1 + 1.0e-6)

    # 6 is a neighbor of 7
    six, seven = algo.items.numbers([6, 7])
    assert matrix[six, seven] > 0

    # and has the correct score
    six_v = simple_ratings[simple_ratings["item_id"] == 6].set_index("user_id").rating
    six_v = six_v - six_v.mean()
    seven_v = simple_ratings[simple_ratings["item_id"] == 7].set_index("user_id").rating
    seven_v = seven_v - seven_v.mean()
    denom = la.norm(six_v.values) * la.norm(seven_v.values)
    six_v, seven_v = six_v.align(seven_v, join="inner")
    num = six_v.dot(seven_v)
    assert matrix[six, seven] == approx(num / denom, 0.01)  # type: ignore


def test_ii_simple_predict():
    history = UserTrainingHistoryLookup()
    history.train(simple_ds)
    algo = ItemKNNScorer(k=30, save_nbrs=500)
    algo.train(simple_ds)

    q = history(3)
    res = algo(q, ItemList([6]))
    _log.info("got predictions: %s", res)
    assert res is not None
    assert len(res) == 1
    assert 6 in res.ids()
    assert not np.isnan(res.scores()[0])


def test_ii_simple_implicit_predict():
    history = UserTrainingHistoryLookup()
    history.train(simple_ds)
    algo = ItemKNNScorer(k=30, feedback="implicit")
    algo.train(from_interactions_df(simple_ratings.loc[:, ["user_id", "item_id"]]))

    q = history(3)
    res = algo(q, ItemList([6]))
    assert res is not None
    assert len(res) == 1
    assert 6 in res.ids()
    assert not np.isnan(res.scores()[0])
    assert np.all(res.scores() > 0)


def test_ii_simple_predict_unknown():
    history = UserTrainingHistoryLookup()
    history.train(simple_ds)
    algo = ItemKNNScorer(k=30, save_nbrs=500)
    algo.train(simple_ds)

    q = history(3)
    res = algo(q, ItemList([6, 100]))
    _log.info("got predictions: %s", res)
    assert res is not None
    assert len(res) == 2
    assert res.ids().tolist() == [6, 100]
    assert not np.isnan(res.scores()[0])
    assert np.isnan(res.scores()[1])


def test_ii_warns_center():
    "Test that item-item warns if you center non-centerable data"
    data = simple_ratings.assign(rating=1)
    algo = ItemKNNScorer(k=5)
    with pytest.warns(DataWarning):
        algo.train(from_interactions_df(data))


@mark.slow
def test_ii_train_ml100k(tmp_path, ml_100k):
    "Test an unbounded model on ML-100K"
    algo = ItemKNNScorer(k=30)
    _log.info("training model")
    algo.train(from_interactions_df(ml_100k))

    _log.info("testing model")

    matrix = algo.sim_matrix.to_scipy()
    assert all(np.logical_not(np.isnan(matrix.data)))
    assert np.all(matrix.data > 0)

    # a little tolerance
    assert np.max(matrix.data) <= 1

    assert algo.item_counts.sum() == len(matrix.data)

    means = ml_100k.groupby("item_id").rating.mean()
    assert means[algo.items.ids()].values == approx(algo.item_means)

    # save
    fn = tmp_path / "ii.mod"
    _log.info("saving model to %s", fn)
    with fn.open("wb") as modf:
        pickle.dump(algo, modf)

    _log.info("reloading model")
    with fn.open("rb") as modf:
        restored = pickle.load(modf)

    r_mat = restored.sim_matrix.to_scipy()

    assert all(r_mat.data > 0)
    assert all(r_mat.data == matrix.data)


@mark.slow
def test_ii_large_models(rng, ml_ratings, ml_ds):
    "Several tests of large trained I-I models"
    _log.info("training limited model")
    MODEL_SIZE = 100
    algo_lim = ItemKNNScorer(k=30, save_nbrs=MODEL_SIZE)
    algo_lim.train(ml_ds)

    _log.info("training unbounded model")
    algo_ub = ItemKNNScorer(k=30)
    algo_ub.train(ml_ds)

    _log.info("testing models")
    mat_lim = algo_lim.sim_matrix.to_scipy()
    assert all(np.logical_not(np.isnan(mat_lim.data)))
    assert mat_lim.data.min() > 0
    # a little tolerance
    assert mat_lim.data.max() <= 1 + 1.0e-6

    means = ml_ratings.groupby("item_id").rating.mean()
    means = means.reindex(algo_ub.items.ids(), fill_value=0.0)
    assert means.values == approx(algo_lim.item_means)

    mat_ub = algo_ub.sim_matrix.to_scipy()
    assert all(np.logical_not(np.isnan(mat_ub.data)))
    assert mat_ub.data.min() > 0
    assert mat_ub.data.max() <= 1 + 1.0e-6

    means = ml_ratings.groupby("item_id").rating.mean()
    means = means.reindex(algo_ub.items.ids(), fill_value=0.0)
    assert means.values == approx(algo_ub.item_means)

    mc_rates = (
        ml_ratings.set_index("item_id")
        .join(pd.DataFrame({"item_mean": means}))
        .assign(rating=lambda df: df.rating - df.item_mean)
    )

    # _log.info("make sure the similarity matrix is sorted")
    # for i in range(algo_lim.items.size):
    #     sp = mat_lim.indptr[i]
    #     ep = mat_lim.indptr[i + 1]
    #     vals = mat_lim.data[sp:ep]
    #     diffs = np.diff(vals)
    #     if np.any(diffs > 0):
    #         _log.error("row %d: %d non-sorted values", i, np.sum(diffs <= 0))
    #         (bad,) = np.nonzero(diffs > 0)
    #         for i in bad:
    #             _log.info("bad indices %d: %d %d", i, vals[i], vals[i + 1])
    #         raise AssertionError(f"{np.sum(diffs <= 0)} non-sorted values")

    _log.info("checking a sample of neighborhoods")
    items = algo_ub.items.ids()
    items = items[algo_ub.item_counts > 0]
    for i in rng.choice(items, 50):
        ipos = algo_ub.items.number(i)
        _log.debug("checking item %d at position %d", i, ipos)
        assert ipos == algo_lim.items.number(i)
        irates = mc_rates.loc[[i], :].set_index("user_id").rating

        ub_row = mat_ub[[ipos]]
        b_row = mat_lim[[ipos]]
        assert len(b_row.data) <= MODEL_SIZE
        ub_cols = ub_row.indices
        b_cols = b_row.indices
        _log.debug("kept %d of %d neighbors", len(b_cols), len(ub_cols))

        # all bounded columns are in the unbounded columns
        _log.debug("checking that bounded columns are a subset of unbounded")
        present = np.isin(b_cols, ub_cols)
        if not np.all(present):
            _log.error("missing items: %s", b_cols[~present])
            _log.error("scores: %s", b_row.data[~present])  # type: ignore
            raise AssertionError(f"missing {np.sum(~present)} values from unbounded")

        # spot-check some similarities
        _log.debug("checking equal similarities")
        for n in rng.choice(ub_cols, min(10, len(ub_cols))):
            n_id = algo_ub.items.id(n)
            n_rates = mc_rates.loc[n_id, :].set_index("user_id").rating
            ir, nr = irates.align(n_rates, fill_value=0)
            cor = ir.corr(nr)
            assert mat_ub[ipos, n] == approx(cor, abs=1.0e-6)

        # short rows are equal
        if len(b_cols) < MODEL_SIZE:
            _log.debug("short row of length %d", len(b_cols))
            assert len(b_row) == len(ub_row)
            assert b_row.data == approx(ub_row.data)
            continue

        # row is truncated - check that truncation is correct
        ub_nbrs = pd.Series(ub_row.data, algo_ub.items.ids(ub_cols))
        b_nbrs = pd.Series(b_row.data, algo_lim.items.ids(b_cols))

        assert len(ub_nbrs) >= len(b_nbrs)
        assert len(b_nbrs) <= MODEL_SIZE
        assert all(b_nbrs.index.isin(ub_nbrs.index))
        # the similarities should be equal!
        b_match, ub_match = b_nbrs.align(ub_nbrs, join="inner")
        assert all(b_match == b_nbrs)
        assert b_match.values == approx(ub_match.values)
        assert b_nbrs.max() == approx(ub_nbrs.max())
        if len(ub_nbrs) > MODEL_SIZE:
            assert len(b_nbrs) == MODEL_SIZE
            ub_shrink = ub_nbrs.nlargest(MODEL_SIZE)
            # the minimums should be equal
            assert ub_shrink.min() == approx(b_nbrs.min())
            # everything above minimum value should be the same set of items
            # the minimum value might be a tie
            ubs_except_min = ub_shrink[ub_shrink > b_nbrs.min()]
            missing = ~ubs_except_min.index.isin(b_nbrs.index)
            if np.any(missing):
                _log.error("missing unbounded values:\n%s", ubs_except_min[missing])
                raise AssertionError(f"missing {np.sum(missing)} unbounded values")


@mark.slow
def test_ii_implicit_large(rng, ml_ratings):
    "Test that implicit-feedback mode works on full test data."
    _log.info("training model")
    NBRS = 5
    NUSERS = 25
    NRECS = 50
    algo = ItemKNNScorer(k=NBRS, feedback="implicit")
    pipe = topn_pipeline(algo)
    pipe.train(from_interactions_df(ml_ratings[["user_id", "item_id"]], item_col="item_id"))

    users = rng.choice(ml_ratings["user_id"].unique(), NUSERS)

    items: Vocabulary = algo.items
    mat: NDArray[np.float32] = algo.sim_matrix.to_scipy().toarray()

    for user in users:
        recs = pipe.run("recommender", query=user, n=NRECS)
        _log.info("user %s recs\n%r", user, recs)
        assert isinstance(recs, ItemList)
        assert len(recs) == NRECS
        urates = ml_ratings[ml_ratings["user_id"] == user]

        smat = mat[items.numbers(urates["item_id"].values), :]
        for row in recs.to_df().itertuples():
            col = smat[:, items.number(row.item_id)]
            top, _is = torch.topk(torch.from_numpy(col), NBRS)
            score = top.sum()
            try:
                assert row.score == approx(score)
            except AssertionError as e:
                _log.error("test failed for user %s item %s", user, row.item_id)
                _log.info("score: %.6f", row.score)
                _log.info("sims:\n%s", col)
                _log.info("total: %.3f", col.sum())
                _log.info("filtered: %s", top)
                _log.info("filtered sum: %.3f", top.sum())
                raise e


def test_ii_known_preds(ml_ds):
    from lenskit import batch

    iknn = ItemKNNScorer(k=20, min_sim=1.0e-6)
    pipe = predict_pipeline(iknn, fallback=False)  # noqa: F821
    _log.info("training %s on ml data", iknn)
    pipe.train(ml_ds)
    _log.info("model means: %s", iknn.item_means)

    dir = Path(__file__).parent
    pred_file = dir / "item-item-preds.csv"
    _log.info("reading known predictions from %s", pred_file)
    known_preds = pd.read_csv(str(pred_file))
    known = ItemListCollection.from_df(known_preds, UserIDKey)

    preds = batch.predict(pipe, known, n_jobs=1)
    preds = preds.to_df().drop(columns=["prediction"], errors="ignore")

    merged = pd.merge(known_preds.rename(columns={"prediction": "expected"}), preds)
    assert len(merged) == len(preds)
    merged["error"] = merged.expected - merged.score
    try:
        assert not any(merged.score.isna() & merged.expected.notna())
    except AssertionError as e:
        bad = merged[merged.score.isna() & merged.expected.notna()]
        _log.error("erroneously missing or present predictions:\n%s", bad)
        raise e

    err = merged.error
    err = err[err.notna()]
    space = np.zeros(7)
    space[1:] = np.logspace(-6, -1, 6)
    counts, edges = np.histogram(np.abs(err), space)
    _log.info("error histogram: %s", counts)
    try:
        # no more than 5 are out-of-bounds
        assert np.sum(space[1:]) < 5
    except AssertionError as e:
        bad = merged[merged.error.notna() & (merged.error.abs() >= 0.01)]
        _log.error("erroneous predictions:\n%s", bad)
        raise e



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import logging
import pickle

import numpy as np
import pandas as pd

from pytest import approx, importorskip, mark

from lenskit.data import Dataset, ItemList, from_interactions_df
from lenskit.metrics import quick_measure_model
from lenskit.testing import BasicComponentTests, ScorerTests

svd = importorskip("lenskit.sklearn.svd")

_log = logging.getLogger(__name__)

simple_df = pd.DataFrame(
    {"item": [1, 1, 2, 3], "user": [10, 12, 10, 13], "rating": [4.0, 3.0, 5.0, 2.0]}
)
simple_ds = from_interactions_df(simple_df)


class TestBiasedSVD(BasicComponentTests, ScorerTests):
    component = svd.BiasedSVDScorer
    config = svd.BiasedSVDConfig(embedding_size=25, damping=10)
    expected_rmse = (0.915, 0.925)

    def verify_models_equivalent(self, orig, copy):
        assert copy.bias_.global_bias == orig.bias_.global_bias
        assert np.all(copy.bias_.user_biases == orig.bias_.user_biases)
        assert np.all(copy.bias_.item_biases == orig.bias_.item_biases)
        assert np.all(copy.user_components_ == orig.user_components_)


def test_svd_basic_build():
    algo = svd.BiasedSVDScorer(features=2)
    algo.train(simple_ds)

    assert algo.user_components_.shape == (3, 2)


def test_svd_predict_basic():
    _log.info("SVD input data:\n%s", simple_df)
    algo = svd.BiasedSVDScorer(features=2, damping=0)
    algo.train(simple_ds)
    _log.info("user means:\n%s", str(algo.bias_.user_biases))
    _log.info("item means:\n%s", str(algo.bias_.item_biases))
    _log.info("user matrix:\n%s", str(algo.user_components_))

    preds = algo(10, ItemList([3]))
    assert len(preds) == 1
    preds = preds.scores("pandas", index="ids")
    assert preds is not None
    assert preds.index[0] == 3
    assert preds.loc[3] >= 0
    assert preds.loc[3] <= 5


def test_svd_predict_bad_item():
    algo = svd.BiasedSVDScorer(features=2)
    algo.train(simple_ds)

    preds = algo(10, ItemList([4]))
    assert len(preds) == 1
    preds = preds.scores("pandas", index="ids")
    assert preds is not None
    assert preds.index[0] == 4
    assert np.isnan(preds.loc[4])


def test_svd_predict_bad_user():
    algo = svd.BiasedSVDScorer(features=2)
    algo.train(simple_ds)

    preds = algo(50, ItemList([3]))
    assert len(preds) == 1
    preds = preds.scores("pandas", index="ids")
    assert preds is not None
    assert preds.index[0] == 3
    assert np.isnan(preds.loc[3])



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from dataclasses import dataclass

import numpy as np
import ray

from lenskit.pipeline import PipelineConfig
from lenskit.splitting import TTSplit

from .spec import TuningSpec


@dataclass
class TuningJobData:
    """
    Data and configuration for a tuning run.
    """

    spec: TuningSpec
    random_seed: np.random.SeedSequence
    data_name: str | None

    factory_ref: ray.ObjectRef | None = None
    data_ref: ray.ObjectRef | None = None

    @property
    def pipeline(self) -> PipelineConfig:
        assert isinstance(self.spec.pipeline, PipelineConfig)
        return self.spec.pipeline

    @property
    def data(self) -> TTSplit:
        assert self.data_ref is not None
        return ray.get(self.data_ref)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from abc import ABC, abstractmethod
from typing import Any, ClassVar, Protocol

import numpy as np
import pyarrow as pa

from lenskit.data import ItemList, ItemListCollection


class MetricFunction(Protocol):
    "Interface for per-list metrics implemented as simple functions."

    @abstractmethod
    def __call__(self, output: ItemList, test: ItemList, /) -> float: ...


class Metric(ABC):
    """
    Base class for LensKit metrics.  Individual metrics need to implement a
    sub-interface, such as :class:`ListMetric` and/or :class:`GlobalMetric`.

    This class defines the interface for metrics. Subclasses should implement
    the `measure_list` method to compute metric values.

    The `summarize()` method has a default implementation that computes the
    mean of the per-list metric values, but subclasses can override it to provide
    more appropriate summary statistics.

    Stability:
        Full

    .. note::

        For simplicity in the analysis code, you cannot simply implement the
        properties of this class on an arbitrary class in order to implement a
        metric with all available behavior such as labeling and defaults;
        you must actually extend this class. This requirement may be relaxed
        in the future.

    The default value to impute when computing statistics over missing values.
    If ``None``, no imputation is done (necessary for metrics like RMSE, where
    the missing value is theoretically infinite).
    """

    @property
    def label(self) -> str:
        """
        The metric's default label in output.
        The base implementation returns the class name by default.
        """
        return self.__class__.__name__

    def __str__(self):
        return f"Metric {self.label}"

    @abstractmethod
    def measure_list(self, output: ItemList, test: ItemList, /) -> Any:
        """
        Compute measurements for a single list.

        Returns:
            - A float for simple metrics
            - Intermediate data for decomposed metrics
            - A dict mapping metric names to values for multi-metric classes
        """
        raise NotImplementedError()  # pragma: no cover

    def extract_list_metrics(self, data: Any, /) -> float | dict[str, float] | None:
        """
        Extract per-list metric(s) from intermediate measurement data.

        Returns:
            - A float for simple metrics
            - A dict mapping metric names to values for multi-metric classes
            - None if no per-list metrics are available
        """
        return None

    @abstractmethod
    def summarize(self, values: list[Any] | pa.Array | pa.ChunkedArray, /) -> dict[str, float]:
        """
        Aggregate intermediate values into summary statistics.

        Returns:
            A dictionary of summary statistics.
        """
        raise NotImplementedError()  # pragma: no cover


class ListMetric(Metric):
    """
    Base class for metrics that measure individual recommendation (or
    prediction) lists, and whose results may be aggregated to compute overall
    metrics.

    For prediction metrics, this is *macro-averaging*.

    Default behavior:
        Implements `summarize()` by averaging per-list results (mean, ignoring NaNs).

    This class implements the Metric interface in terms of the measure_list method.

    Stability:
        Full
    """

    default: ClassVar[float | None] = 0.0

    @abstractmethod
    def measure_list(self, output: ItemList, test: ItemList, /) -> float:
        """
        Compute the metric value for a single result list.

        Individual metric classes need to implement this method.
        """
        raise NotImplementedError()  # pragma: no cover

    def extract_list_metrics(self, data: Any, /) -> float:
        """
        Return the given per-list metric result.
        """
        return data

    def summarize(
        self, values: list[Any] | pa.Array | pa.ChunkedArray, /
    ) -> dict[str, float | None]:
        """
        Summarize per-list metric values

        Returns:
            A dictionary containing mean, median, and std.
        """
        if isinstance(values, (pa.Array, pa.ChunkedArray)):
            values = values.to_pylist()

        numeric_values = [
            float(v) for v in values if isinstance(v, (int, float, np.integer, np.floating))
        ]

        if not numeric_values:
            return {"mean": None, "median": None, "std": None}

        arr = np.array(numeric_values, dtype=np.float64)
        return {
            "mean": float(np.mean(arr)),
            "median": float(np.median(arr)),
            "std": float(np.std(arr, ddof=1)) if len(arr) > 1 else 0.0,
        }


class GlobalMetric(Metric):
    """
    Base class for metrics that measure entire runs at a time.

    For prediction metrics, this is *micro-averaging*.

    Stability:
        Full
    """

    @abstractmethod
    def measure_run(self, output: ItemListCollection, test: ItemListCollection, /) -> float:
        """
        Compute a metric value for an entire run.

        Individual metric classes need to implement this method.
        """
        raise NotImplementedError()  # pragma: no cover

    def measure_list(self, output: ItemList, test: ItemList, /) -> Any:
        raise NotImplementedError("Global metrics don't support per-list measurement")

    def summarize(self, values: list[Any] | pa.Array | pa.ChunkedArray, /) -> float:
        raise NotImplementedError("Global metrics should implement measure_run instead")


class DecomposedMetric(Metric):
    """
    Deprecated base class for decomposed metrics.

    .. deprecated:: 2025.4
        This class is deprecated and its functionality has been moved to :class:`Metric`.
        It is scheduled for removal in 2026.

    Base class for metrics that measure entire runs through flexible
    aggregations of per-list intermediate measurements.  They can optionally
    extract individual-list metrics from the per-list measurements.

    Stability:
        Full
    """

    def measure_list(self, output: ItemList, test: ItemList, /) -> Any:
        return self.compute_list_data(output, test)

    def extract_list_metrics(self, data: Any, /) -> float | None:
        return self.extract_list_metric(data)

    def summarize(self, values: list[Any] | pa.Array | pa.ChunkedArray, /) -> dict[str, float]:
        if isinstance(values, (pa.Array, pa.ChunkedArray)):
            values = values.to_pylist()
        result = self.global_aggregate(values)
        if isinstance(result, (float, int, np.floating, np.integer)):
            return {"value": float(result)}
        return result

    @abstractmethod
    def compute_list_data(self, output: ItemList, test: ItemList, /) -> Any:
        """
        Compute measurements for a single list.

        Use `measure_list` in `Metric` for new implementations.
        """
        raise NotImplementedError()  # pragma: no cover

    def extract_list_metric(self, data: Any, /) -> float | None:
        """
        Extract a single-list metric from the per-list measurement result (if
        applicable).

        Returns:
            The per-list metric, or ``None`` if this metric does not compute
            per-list metrics.

        Implement :meth:`Metric.extract_list_metrics` in new implementations.
        """
        return None

    @abstractmethod
    def global_aggregate(self, values: list[Any], /) -> float | dict[str, float]:
        """
        Aggregate list metrics to compute a global value.

        Implement :meth:`Metric.summarize` in new implementations.
        """
        raise NotImplementedError()  # pragma: no cover



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

# pyright: strict
from pytest import fail, raises

from lenskit.pipeline import Lazy, PipelineBuilder


def fallback(first: int | None, second: Lazy[int]) -> int:
    if first is not None:
        return first
    else:
        return second.get()


def test_lazy_input():
    pipe = PipelineBuilder()
    a = pipe.create_input("a", int)
    b = pipe.create_input("b", int)

    def negative(x: int) -> int:
        return -x

    def double(x: int) -> int:
        return x * 2

    def add(x: int, y: int) -> int:
        return x + y

    nd = pipe.add_component("double", double, x=a)
    nn = pipe.add_component("negate", negative, x=a)
    fb = pipe.add_component("fill-operand", fallback, first=b, second=nn)
    na = pipe.add_component("add", add, x=nd, y=fb)

    pipe = pipe.build()
    # 3 * 2 + -3 = 3
    assert pipe.run(na, a=3) == 3


def test_lazy_only_run_if_needed():
    pipe = PipelineBuilder()
    a = pipe.create_input("a", int)
    b = pipe.create_input("b", int)

    def negative(x: int) -> int:
        fail("fallback component run when not needed")

    def double(x: int) -> int:
        return x * 2

    def add(x: int, y: int) -> int:
        return x + y

    nd = pipe.add_component("double", double, x=a)
    nn = pipe.add_component("negate", negative, x=a)
    fb = pipe.add_component("fill-operand", fallback, first=b, second=nn)
    na = pipe.add_component("add", add, x=nd, y=fb)

    pipe = pipe.build()
    assert pipe.run(na, a=3, b=8) == 14


def test_lazy_fail_with_missing_options():
    pipe = PipelineBuilder()
    a = pipe.create_input("a", int)
    b = pipe.create_input("b", int)

    def negative(x: int) -> int | None:
        return None

    def double(x: int) -> int:
        return x * 2

    def add(x: int, y: int) -> int:
        return x + y

    nd = pipe.add_component("double", double, x=a)
    nn = pipe.add_component("negate", negative, x=a)
    fb = pipe.add_component("fill-operand", fallback, first=b, second=nn)
    na = pipe.add_component("add", add, x=nd, y=fb)

    pipe = pipe.build()
    with raises(TypeError):
        pipe.run(na, a=3)


def test_lazy_transitive():
    "test that a fallback works if a dependency's dependency fails"
    pipe = PipelineBuilder()
    ia = pipe.create_input("a", int)
    ib = pipe.create_input("b", int)

    def double(x: int) -> int:
        return 2 * x

    # two components, each with a different input
    c1 = pipe.add_component("double-a", double, x=ia)
    c2 = pipe.add_component("double-b", double, x=ib)
    # use the first that succeeds
    c = pipe.add_component("fill-operand", fallback, first=c1, second=c2)

    pipe = pipe.build()
    # omitting the first input should result in the second component
    assert pipe.run(c, b=17) == 34


def test_lazy_transitive_deeper():
    "deeper transitive fallback test"
    pipe = PipelineBuilder()
    a = pipe.create_input("a", int)
    b = pipe.create_input("b", int)

    def negative(x: int) -> int:
        return -x

    def double(x: int) -> int:
        return x * 2

    nd = pipe.add_component("double", double, x=a)
    nn = pipe.add_component("negate", negative, x=nd)
    nr = pipe.add_component("fill-operand", fallback, first=nn, second=b)

    pipe = pipe.build()
    assert pipe.run(nr, b=8) == 8


def test_lazy_transitive_nodefail():
    "deeper transitive fallback test"
    pipe = PipelineBuilder()
    a = pipe.create_input("a", int)
    b = pipe.create_input("b", int)

    def negative(x: int) -> int | None:
        # make this return None in some cases to trigger failure
        if x >= 0:
            return -x
        else:
            return None

    def double(x: int) -> int:
        return x * 2

    nd = pipe.add_component("double", double, x=a)
    nn = pipe.add_component("negate", negative, x=nd)
    nr = pipe.add_component("fill-operand", fallback, first=nn, second=b)

    pipe = pipe.build()
    assert pipe.run(nr, a=2, b=8) == -4
    assert pipe.run(nr, a=-7, b=8) == 8



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from lenskit.pipeline.config import PipelineInput
from lenskit.pipeline.nodes import InputNode


def test_untyped_input():
    node = InputNode("scroll")

    cfg = PipelineInput.from_node(node)
    print(cfg)
    assert cfg.name == "scroll"
    assert cfg.types is None


def test_input_with_type():
    node = InputNode("scroll", types={str})

    cfg = PipelineInput.from_node(node)
    print(cfg)
    assert cfg.name == "scroll"
    assert cfg.types == {"str"}


def test_input_with_none():
    node = InputNode("scroll", types={str, type(None)})

    cfg = PipelineInput.from_node(node)
    print(cfg)
    assert cfg.name == "scroll"
    assert cfg.types == {"None", "str"}


def test_input_with_generic():
    node = InputNode("scroll", types={list[str]})

    cfg = PipelineInput.from_node(node)
    print(cfg)
    assert cfg.name == "scroll"
    assert cfg.types == {"list"}



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
LensKit general configuration
"""

from __future__ import annotations

import json
import tomllib
import warnings
from os import PathLike
from pathlib import Path
from typing import Annotated

from annotated_types import Gt, Le
from pydantic import BaseModel, JsonValue
from pydantic_settings import BaseSettings, SettingsConfigDict, TomlConfigSettingsSource
from typing_extensions import Any, TypedDict, TypeVar, overload

from lenskit.diagnostics import ConfigWarning
from lenskit.logging import get_logger
from lenskit.random import init_global_rng

__all__ = [
    "lenskit_config",
    "configure",
    "LenskitSettings",
    "RandomSettings",
    "MachineSettings",
    "PowerQueries",
    "PrometheusSettings",
]

M = TypeVar("M", bound=BaseModel)
"Model class for general configuration loading."
SettingsClass = TypeVar("SettingsClass", bound="LenskitSettings", default="LenskitSettings")
_log = get_logger(__name__)
_settings: LenskitSettings | None = None


def lenskit_config() -> LenskitSettings:
    """
    Get the LensKit configuration.

    If no configuration has been specified, returns a default settings object.
    """
    global _settings

    if _settings is None:
        return LenskitSettings()
    else:
        return _settings


class PowerQueries(TypedDict, total=False):
    """
    Queries for requesting power consumption data from Prometheus.

    Each entry is a Python format string (for :meth:`str.format`), that will be
    used to format a dictionary ``{"elapsed": time_ms}``.  The result should be
    a valid Prometheus query that returns the power consumption in Joules (watt
    seconds) over the last ``time_ms`` milliseconds.
    """

    system: str
    "Total (chassis or system) power consumption."
    cpu: str
    "CPU power consumption."
    gpu: str
    "GPU power consumption."


class RandomSettings(BaseModel):
    """
    Random number generator configuration.
    """

    seed: int | None = None
    """
    The root RNG seed.
    """


class MachineSettings(BaseModel, extra="allow"):
    """
    Definition for a single machine.

    A “machine” is a computer (or cluster) that is in use for running LensKit
    experiments.  Many users won't use this, but if you want to use the power
    consumption monitoring, you will need to define how to measure power for the
    different machines in use.
    """

    description: str | None = None
    "Short description for this machine."
    power_queries: PowerQueries = {}
    "Prometheus queries to collect power metrics for this machine."


class PrometheusSettings(BaseModel):
    """
    Prometheus configuration settings.

    LensKit's task logging supports querying Prometheus for task-related metrics
    such as power consumption.
    """

    url: str | None = None


class TuneSettings(BaseModel):
    """
    LensKit hyperparameter tuning settings.
    """

    jobs: int | None = None
    """
    Number of allowed hyperparameter tuning jobs.
    """
    gpu_mult: Annotated[float, Gt(0), Le(1.0)] = 1.0
    """
    Multiplier for tuning job GPU requirements.  This is to coarsely adapt GPU
    requirements from configuration files to the local machine.  If a tuning
    specificataion requires 1 GPU, but your machine has enough capacity to run
    two jobs in parallel on a single GPU, you can set this to 0.5 to modify the
    tuning jobs to require 0.5 GPUs each.
    """


class LenskitSettings(BaseSettings, extra="allow"):
    """
    Definition of LensKit settings.

    LensKit supports loading various settings from configuration files
    and the environment as a consistent way to control LensKit's various
    control surfaces.

    Stability:
        Experimental
    """

    model_config = SettingsConfigDict(
        nested_model_default_partial_update=True, env_prefix="LK_", env_nested_delimiter="__"
    )

    random: RandomSettings = RandomSettings()
    """
    Random number generator configuration.
    """

    machine: str | None = None
    """
    The name of the machine running experiments.

    This is usually set in ``lenskit.local.toml`` or the ``LK_MACHINE`` environment
    variable.
    """
    prometheus: PrometheusSettings = PrometheusSettings()
    """
    Prometheus settings for task metric collection.
    """
    machines: dict[str, MachineSettings] = {}
    """
    Description of different machines used in the experiment(s), to support
    things like collecting power metrics.
    """

    tune: TuneSettings = TuneSettings()
    """
    LensKit tuning settings.
    """

    @property
    def current_machine(self) -> MachineSettings | None:
        if self.machine:
            if ms := self.machines.get(self.machine, None):
                return ms

    @classmethod
    def settings_customise_sources(
        cls, settings_cls, init_settings, env_settings, dotenv_settings, file_secret_settings
    ):
        return (
            init_settings,
            env_settings,
            dotenv_settings,
            file_secret_settings,
        )


@overload
def configure(cfg_dir: Path | None = None) -> LenskitSettings: ...
@overload
def configure(
    cfg_dir: Path | None = None, *, settings_cls: type[SettingsClass]
) -> SettingsClass: ...
def configure(
    cfg_dir: Path | None = None, *, settings_cls=LenskitSettings, _set_global: bool = True
) -> Any:
    """
    Initialize LensKit configuration.

    LensKit does **not** automatically read configuration files — if this
    function is never called, then configuration will entirely be done through
    defaults and environment varibles.

    This function will automatically configure the global RNG, if a seed is
    specified.  It does **not** configure logging.

    Args:
        cfg_dir:
            The directory in which to look for configuration files..  If not
            provided, uses the current directory.
        settings_cls:
            The base LensKit settings class.  Rarely used, only needed if a
            project wants to to extend LensKit settings with their own settings.

    Returns:
        The configured LensKit settings.
    """
    global _settings

    if _settings is not None and _set_global:
        warnings.warn("LensKit already configured, overwriting configuration", ConfigWarning)

    # define a subclass so we can specify the configuration location
    toml_files = ["lenskit.toml", "lenskit.local.toml"]
    if cfg_dir is not None:
        toml_files = [cfg_dir / f for f in toml_files]

    class LenskitFileSettings(settings_cls):
        @classmethod
        def settings_customise_sources(
            cls, settings_cls, init_settings, env_settings, dotenv_settings, file_secret_settings
        ):
            return (
                init_settings,
                env_settings,
                dotenv_settings,
                file_secret_settings,
                TomlConfigSettingsSource(
                    settings_cls,
                    cfg_dir / "lenskit.local.toml" if cfg_dir is not None else "lenskit.local.toml",
                ),
                TomlConfigSettingsSource(
                    settings_cls,
                    cfg_dir / "lenskit.toml" if cfg_dir is not None else "lenskit.toml",
                ),
            )

    settings = LenskitFileSettings()
    if _set_global:
        _settings = settings
        if settings.random.seed is not None:
            init_global_rng(settings.random.seed)

    return settings


def locate_configuration_root(
    *,
    cwd: Path | str | PathLike[str] | None = None,
    abort_at_pyproject: bool = True,
    abort_at_gitroot: bool = True,
) -> Path | None:
    """
    Search for a configuration root containing a ``lenskit.toml`` file.

    This searches for a ``lenskit.toml`` file, beginning in the current working
    directory (or the alternate ``cwd`` if provided), and searching upward until
    one is found.  Search stops if a ``pyproject.toml`` file or ``.git``
    directory is found without encountering ``lenskit.toml``.
    """

    if cwd is None:
        cwd = Path()
    elif not isinstance(cwd, Path):
        cwd = Path(cwd)
    cwd = cwd.resolve()

    log = _log.bind(cwd=str(cwd))
    log.debug("searching for lenskit.toml")
    while cwd is not None:
        log.debug("checking if lenskit.toml exists", dir=str(cwd))
        if (cwd / "lenskit.toml").exists():
            return cwd

        if abort_at_pyproject and (cwd / "pyproject.toml").exists():
            break

        if abort_at_gitroot and (cwd / ".git").exists():
            break

        if cwd.parent == cwd:
            break
        else:
            cwd = cwd.parent


@overload
def load_config_data(path: Path | PathLike[str], model: None = None) -> JsonValue: ...
@overload
def load_config_data(path: Path | PathLike[str], model: type[M]) -> M: ...
def load_config_data(path: Path | PathLike[str], model: type[M] | None = None):
    """
    General-purpose function to automatically load configuration data and
    optionally validate with a model.

    Args:
        path:
            The path to the configuration file.
        model:
            The Pydantic model class to validate.
    """
    path = Path(path)
    text = path.read_text()

    match path.suffix:
        case ".json" if model is not None:
            return model.model_validate_json(text)
        case ".json":
            data = json.loads(text)
        case ".toml":
            data = tomllib.loads(text)
        case ".yaml" | ".yml":
            import yaml

            data = yaml.load(text, yaml.SafeLoader)

        case _:
            raise ValueError(f"unsupported configuration type for {path}")

    if model is None:
        return data
    else:
        return model.model_validate(data)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
LensKit dataset abstraction.
"""

# pyright: basic
from __future__ import annotations

import functools
import io
from abc import abstractmethod
from collections.abc import Callable
from os import PathLike

import pandas as pd
import pyarrow as pa
import scipy.sparse as sps
import torch
from humanize import metric
from numpy.typing import NDArray
from typing_extensions import Any, Literal, TypeAlias, TypedDict, TypeVar, overload

from lenskit.diagnostics import DataError
from lenskit.logging import get_logger

from .container import DataContainer
from .entities import EntitySet
from .items import ItemList
from .matrix import CSRStructure
from .relationships import MatrixRelationshipSet, RelationshipSet
from .schema import DataSchema, id_col_name
from .types import ID, LAYOUT
from .vocab import Vocabulary

_log = get_logger(__name__)

ACTION_FIELDS: TypeAlias = Literal["ratings", "timestamps"] | str

K = TypeVar("K")


def _uses_data(func):
    """
    Decorator to make sure the data is loaded.
    """

    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        self._ensure_loaded()
        return func(self, *args, **kwargs)

    return wrapper


class DatasetState(TypedDict):
    data: DataContainer


class Dataset:
    """
    Representation of a data set for LensKit training, evaluation, etc. Data can
    be accessed in a variety of formats depending on the needs of a component.
    See :ref:`data-model` for details of the LensKit data model.

    Dataset objects should not be directly constructed; instead, use a
    :class:`DatasetBuilder`, :meth:`load`, or :func:`from_interactions_df`.

    .. note::

        Zero-copy conversions are used whenever possible, so client code **must
        not** modify returned data in-place.

    Args:
        data:
            The container for this dataset's data, or a function that will
            return such a container to create a lazy-loaded dataset.

    .. stability:: caller
    """

    _data: DataContainer
    _data_thunk: Callable[[], DataContainer | Dataset]
    _entities: dict[str, EntitySet]
    _relationships: dict[str, RelationshipSet]

    def __init__(self, data: DataContainer | Callable[[], DataContainer | Dataset]):
        if isinstance(data, DataContainer):
            self._data = data
            self._init_caches()
        else:
            self._data_thunk = data

    @classmethod
    def load(cls, path: str | PathLike[str]) -> Dataset:
        """
        Load a dataset in the LensKit native format.

        Args:
            path:
                The path to the dataset to load.

        Returns:
            The loaded dataset.
        """
        container = DataContainer.load(path)
        return cls(container)

    def save(self, path: str | PathLike[str]):
        """
        Save the data set in the LensKit native format.

        Args:
            path:
                The path in which to save the data set (will be created as a
                directory).
        """
        self._data.save(path)

    @property
    def name(self) -> str | None:
        """
        Get the dataset's name.
        """
        return self.schema.name

    def _ensure_loaded(self):
        if not hasattr(self, "_data"):
            _log.debug("lazy-loading dataset")
            data = self._data_thunk()
            del self._data_thunk
            if isinstance(data, DataContainer):
                self._data = data
                self._init_caches()
            elif isinstance(data, Dataset):
                data._ensure_loaded()
                self._data = data._data
                self._entities = data._entities
                self._relationships = data._relationships
            else:  # pragma: nocover
                raise TypeError("invalid thunk return: " + str(type(data)))

    def _init_caches(self):
        "Initialize internal caches for this dataset."
        log = _log.bind(dataset=self.name)
        self._data.normalize()

        self._entities = {}
        self._relationships = {}

        log.debug("initializing dataset caches")

        for name, schema in self._data.schema.entities.items():
            tbl = self._data.tables[name]
            id_name = id_col_name(name)
            ids = tbl.column(id_name)
            log.debug("creating entity vocabulary", entity=name)
            vocab = Vocabulary(ids, name=name, reorder=False)
            log.debug("creating entity set", entity=name)
            self._entities[name] = EntitySet(name, schema, vocab, tbl)

        for name, schema in self._data.schema.relationships.items():
            tbl = self._data.tables[name]
            relationship_vocab = {e: self.entities(e).vocabulary for e in schema.entities}
            if not schema.repeats.is_present and len(schema.entities) == 2:
                log.debug("creating matrix relationship set", relationship=name)
                self._relationships[name] = MatrixRelationshipSet(
                    name, relationship_vocab, schema, tbl
                )
            else:
                log.debug("creating relationship set", relationship=name)
                self._relationships[name] = RelationshipSet(name, relationship_vocab, schema, tbl)

    @property
    @_uses_data
    def schema(self) -> DataSchema:
        """
        Get the schema of this dataset.
        """
        return self._data.schema

    @property
    @_uses_data
    def items(self) -> Vocabulary:
        """
        The items known by this dataset.
        """
        return self.entities("item").vocabulary

    @property
    @_uses_data
    def users(self) -> Vocabulary:
        """
        The users known by this dataset.
        """
        return self.entities("user").vocabulary

    @property
    def item_count(self) -> int:
        return len(self.items)

    @property
    def user_count(self) -> int:
        return len(self.users)

    @_uses_data
    def entities(self, name: str) -> EntitySet:
        """
        Get the entities of a particular type / class.
        """
        eset = self._entities.get(name, None)
        if eset is None:
            raise DataError(f"entity class {name} is not defined")
        return eset

    @_uses_data
    def relationships(self, name: str) -> RelationshipSet:
        """
        Get the relationship records of a particular type / class.
        """
        rset = self._relationships.get(name, None)
        if rset is None:
            raise DataError(f"relationship class {name} is not defined")

        return rset

    @_uses_data
    def interactions(self, name: str | None = None) -> RelationshipSet:
        """
        Get the interaction records of a particular class.  If no class is
        specified, returns the default interaction class.
        """
        if name is None:
            name = self.default_interaction_class()
        rels = self.relationships(name)
        if not rels.is_interaction:
            raise DataError(f"relationship class {name} is not an interaction class")
        return rels

    def default_interaction_class(self) -> str:
        schema = self.schema
        if schema.default_interaction:
            return schema.default_interaction

        i_classes = [name for (name, rs) in schema.relationships.items() if rs.interaction]
        if len(i_classes) == 1:
            return i_classes[0]
        else:
            raise RuntimeError("no default interaction class specified")

    @property
    def interaction_count(self) -> int:
        """
        Count the total number of interactions of the default class, taking into
        account any ``count`` attribute.
        """
        return self.interactions().count()

    @overload
    @abstractmethod
    def interaction_table(
        self,
        *,
        format: Literal["pandas"],
        fields: str | list[str] | None = None,
        original_ids: bool = False,
    ) -> pd.DataFrame: ...
    @overload
    @abstractmethod
    def interaction_table(
        self, *, format: Literal["numpy"], fields: str | list[str] | None = None
    ) -> dict[str, NDArray[Any]]: ...
    @overload
    @abstractmethod
    def interaction_table(
        self, *, format: Literal["arrow"], fields: str | list[str] | None = None
    ) -> pa.Table: ...
    @abstractmethod
    @_uses_data
    def interaction_table(
        self,
        *,
        format: str,
        fields: str | list[str] | None = None,
        original_ids: bool = False,
    ) -> Any:
        """
        Get the user-item interactions as a table in the requested format. The
        table is not in a specified order.  Interactions may be repeated (e.g.
        the same user may listen to a song multiple times).  For a non-repeated
        “ratings matrix” view of the data, see :meth:`interaction_matrix`.

        This is a convenince wrapper on top of :meth:`interactions` and the
        methods of :class:`RelationshipSet`.

        .. warning::
            Client code **must not** perform in-place modifications on the table
            returned from this method.  Whenever possible, it will be a shallow
            view on top of the underlying storage, and modifications may corrupt
            data for other code.

        Args:
            format:
                The desired data format.  Currently-supported formats are:

                * ``"pandas"`` — returns a :class:`pandas.DataFrame`.  The index
                  is not meaningful.
                * ``"arrow"`` — returns a PyArrow :class:`~pa.Table`.  The index
                  is not meaningful.
                * ``"numpy"`` — returns a dictionary mapping names to arrays.
            fields:
                Which fields (attributes) to include, or ``None`` to include all
                fields. Commonly-available fields include ``"rating"`` and
                ``"timestamp"``.
            original_ids:
                If ``True``, return user and item IDs as represented in the
                original source data in columns named ``user_id`` and
                ``item_id``, instead of the user and item numbers typically
                returned.

        Returns:
            The user-item interaction log in the specified format.
        """
        iset = self.interactions()
        if format == "pandas":
            return iset.pandas(attributes=fields, ids=original_ids)
        else:
            table = iset.arrow(attributes=fields, ids=original_ids)
            if format == "numpy":
                return {c: table.column(c).to_numpy() for c in table.column_names}
            elif format == "arrow":
                return table
            else:
                raise ValueError(f"unsupported format {format}")

    @overload
    @abstractmethod
    def interaction_matrix(
        self,
        *,
        format: Literal["pandas"],
        field: str | None = None,
        original_ids: bool = False,
    ) -> pd.DataFrame: ...
    @overload
    @abstractmethod
    def interaction_matrix(
        self,
        *,
        format: Literal["torch"],
        layout: Literal["csr", "coo"] = "csr",
        field: str | None = None,
    ) -> torch.Tensor: ...
    @overload
    @abstractmethod
    def interaction_matrix(
        self,
        *,
        format: Literal["scipy"],
        layout: Literal["coo"],
        field: str | None = None,
    ) -> sps.coo_array: ...
    @overload
    @abstractmethod
    def interaction_matrix(
        self,
        *,
        format: Literal["scipy"],
        layout: Literal["csr"] = "csr",
        field: str | None = None,
    ) -> sps.csr_array: ...
    @overload
    @abstractmethod
    def interaction_matrix(
        self,
        *,
        format: Literal["structure"],
        layout: Literal["csr"] = "csr",
    ) -> CSRStructure: ...
    @abstractmethod
    @_uses_data
    def interaction_matrix(
        self,
        *,
        format: str,
        layout: LAYOUT = "csr",
        field: str | None = None,
        original_ids: bool = False,
        legacy: bool = False,
    ) -> Any:
        """
        Get the user-item interactions as “ratings” matrix from the default
        interaction class.  Interactions are not repeated, and are coalesced
        with the default coalescing strategy for each attribute.

        The matrix may be returned in “coordinate” format, in which case it is
        comparable to :meth:`interaction_table` but without repeated
        interactions, or it may be in a compressed sparse row format.

        This is a convenince wrapper on top of :meth:`interactions` and the
        methods of :class:`MatrixRelationshipSet`.

        .. warning::

            Client code **must not** perform in-place modifications on the
            matrix returned from this method.  Whenever possible, it will be a
            shallow view on top of the underlying storage, and modifications may
            corrupt data for other code.

        Args:
            format:
                The desired data format.  Currently-supported formats are:

                * ``"pandas"`` — returns a :class:`pandas.DataFrame`.
                * ``"torch"`` — returns a sparse :class:`torch.Tensor` (see
                  :mod:`torch.sparse`).
                * ``"scipy"`` — returns a sparse array from :mod:`scipy.sparse`.
                * ``"structure"`` — returns a :class:`~matrix.CSRStructure`
                  containing only the user and item numbers in compressed sparse
                  row format.
            field:
                Which field to return in the matrix.  Common fields include
                ``"rating"`` and ``"timestamp"``.

                If unspecified (``None``), this will yield an implicit-feedback
                indicator matrix, with 1s for observed items, except for the
                ``"pandas"`` format, which will return all attributes.  Specify
                an empty list to return a Pandas data frame with only the user
                and item attributes.
            layout:
                The layout for a sparse matrix.  Can be either ``csr`` or
                ``coo``, or ``None`` to use the default for the specified
                format.  Ignored for the Pandas format.
            original_ids:
                ``True`` to return user and item IDs instead of numbers in a
                ``pandas``-format matrix.
        """
        iset = self.interactions().matrix()

        match format:
            case "pandas":
                return iset.pandas(attributes=field, ids=original_ids)
            case "scipy":
                return iset.scipy(attribute=field, layout=layout, legacy=legacy)
            case "torch":
                return iset.torch(attribute=field, layout=layout)
            case "structure":
                if layout != "csr":
                    raise ValueError(f"unsupported layout {layout} for CSR structure")
                if field is not None:
                    raise ValueError("structure does not support fields")
                return iset.csr_structure()
            case _:
                raise ValueError(f"unknown matrix format {format}")

    @abstractmethod
    @overload
    def user_row(self, user_id: ID) -> ItemList | None: ...
    @abstractmethod
    @overload
    def user_row(self, *, user_num: int) -> ItemList: ...
    @abstractmethod
    def user_row(
        self, user_id: ID | None = None, *, user_num: int | None = None
    ) -> ItemList | None:
        """
        Get a user's row from the interaction matrix for the default interaction
        class, using :ref:`default coalsecing <coalescing-defaults>` for
        repeated interactions.  Available fields are returned as fields. If the
        dataset has ratings, these are provided as a ``rating`` field, **not**
        as the item scores.  The item list is unordered, but items are returned
        in order by item number.

        Args:
            user_id:
                The ID of the user to retrieve.
            user_num:
                The number of the user to retrieve.

        Returns:
            The user's interaction matrix row, or ``None`` if no user with that
            ID exists.
        """
        iset = self.interactions().matrix()
        return iset.row_items(id=user_id, number=user_num)

    def item_stats(self) -> pd.DataFrame:
        """
        Get item statistics from the default interaction class.

        Returns:
            A data frame indexed by item ID with the interaction statistics. See
            :ref:`interaction-stats` for a description of the columns returned.

            The index is aligned with the vocabulary, so ``iloc`` works with item
            numbers.
        """
        iset = self.interactions().matrix()
        if iset.col_type != "item":
            raise RuntimeError("default interactions do not have item columns")
        return iset.col_stats()

    def user_stats(self) -> pd.DataFrame:
        """
        Get user statistics from the default interaction class.

        Returns:
            A data frame indexed by user ID with the interaction statistics. See
            :ref:`interaction-stats` for a description of the columns returned.

            The index is the vocabulary, so ``iloc`` works with user numbers.
        """
        iset = self.interactions().matrix()
        if iset.row_type != "user":
            raise RuntimeError("default interactions do not have user columns")
        return iset.row_stats()

    def __getstate__(self) -> DatasetState:
        self._ensure_loaded()
        return {
            "data": self._data,
        }

    def __setstate__(self, state: DatasetState):
        self._data = state["data"]
        self._init_caches()

    def __str__(self) -> str:
        s = "<Dataset"
        if self.name is not None:
            s += " " + self.name
        s += " ({} users, {} items, {} interactions)>".format(
            metric(self.user_count), metric(self.item_count), metric(self.interaction_count)
        )
        return s

    def __repr__(self) -> str:
        out = io.StringIO()
        out.write("<Dataset")
        if self.name is not None:
            out.write(" " + self.name)
        out.write(" {\n")
        for entity in self.schema.entities:
            eset = self._entities.get(entity, None)
            if eset is not None:
                out.write("  {}: {:,d},\n".format(entity, eset.count()))
            else:
                out.write("  {}: <not loaded>,\n".format(entity))
        for rel in self.schema.relationships:
            rset = self._relationships.get(rel, None)
            if rset is not None:
                out.write("  {}: {:,d},\n".format(rel, rset.count()))
            else:
                out.write("  {}: <not loaded>,\n".format(rel))
        out.write("}")

        return out.getvalue()



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from __future__ import annotations

import torch
from torch.nn import functional as F
from typing_extensions import Literal, override

from lenskit.data import Dataset
from lenskit.flexmf._model import FlexMFModel

from ._base import FlexMFConfigBase, FlexMFScorerBase
from ._training import (
    FlexMFTrainerBase,
    FlexMFTrainingBatch,
    FlexMFTrainingData,
)


class FlexMFExplicitConfig(FlexMFConfigBase):
    """
    Configuration for :class:`FlexMFExplicitScorer`.  This class overrides
    certain base class defaults for better explicit-feedback performance.

    Stability:
        Experimental
    """

    regularization: float = 0.1
    reg_method: Literal["AdamW", "L2"] | None = "L2"


class FlexMFExplicitScorer(FlexMFScorerBase):
    """
    Explicit-feedback rating prediction with FlexMF.  This realizes a biased
    matrix factorization model (similar to :class:`lenskit.als.BiasedMF`)
    trained with PyTorch.

    Stability:
        Experimental
    """

    global_bias: float
    config: FlexMFExplicitConfig

    def create_trainer(self, data, options):
        return FlexMFExplicitTrainer(self, data, options)

    def score_items(self, users: torch.Tensor, items: torch.Tensor) -> torch.Tensor:
        return super().score_items(users, items) + self.global_bias


class FlexMFExplicitTrainer(FlexMFTrainerBase[FlexMFExplicitScorer]):
    @override
    def prepare_data(self, data: Dataset) -> FlexMFTrainingData:
        """
        Set up the training data and context for the scorer.
        """

        ratings = data.interaction_matrix(field="rating", format="torch", layout="coo")

        # extract the components: row, column, value.
        # indices are a matrix in Torch COO format.
        indices = ratings.indices()
        rm_users = indices[0, :]
        rm_items = indices[1, :]
        rm_values = ratings.values().to(torch.float32)

        # compute the global mean (global bias), and subtract from all rating values
        mean = rm_values.mean()
        rm_values = rm_values - mean

        # save data we learned at this stage
        self.component.global_bias = mean.item()
        self.component.users = data.users
        self.component.items = data.items

        return FlexMFTrainingData(
            batch_size=self.config.batch_size,
            n_users=data.user_count,
            n_items=data.item_count,
            users=rm_users,
            items=rm_items,
            fields={"ratings": rm_values},
        ).to(self.device)

    @override
    def create_model(self) -> FlexMFModel:
        """
        Prepare the model for training.
        """
        return FlexMFModel(
            self.config.embedding_size,
            self.data.n_users,
            self.data.n_items,
            self.torch_rng,
            sparse=self.config.reg_method != "AdamW",
            init_scale=0.1,
        )

    @override
    def train_batch(self, batch: FlexMFTrainingBatch) -> torch.Tensor:
        if self.explicit_norm:
            result = self.call_model(batch.users, batch.items, return_norm=True)
            pred = result[0, :]
            norm = torch.mean(result[1, :]) * self.config.regularization
        else:
            pred = self.call_model(batch.users, batch.items, return_norm=False)
            norm = 0.0

        loss = F.mse_loss(pred, batch.fields["ratings"])

        loss_all = loss + norm

        loss_all.backward()
        self.opt.step()

        return loss



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import numpy as np
import pyarrow as pa
import scipy.sparse as sps
import torch

import hypothesis.extra.numpy as nph
import hypothesis.strategies as st
from hypothesis import HealthCheck, given, note, settings
from pytest import approx, mark, skip

from lenskit.torch import safe_tensor


@given(
    nph.arrays(
        st.one_of(nph.floating_dtypes(endianness="="), nph.integer_dtypes(endianness="=")),
        nph.array_shapes(max_dims=3),
    )
)
def test_safe_tensor_from_numpy(arr):
    t = safe_tensor(arr)
    assert torch.is_tensor(t)

    assert t.shape == arr.shape


@given(
    nph.arrays(
        st.one_of(nph.floating_dtypes(endianness="="), nph.integer_dtypes(endianness="=")),
        nph.array_shapes(max_dims=1),
    )
)
def test_safe_tensor_from_arrow(arr):
    arrow = pa.array(arr)
    t = safe_tensor(arrow)
    assert torch.is_tensor(t)

    assert t.shape == arr.shape



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Batch-run recommendation pipelines for evaluation.
"""

from __future__ import annotations

from collections.abc import Iterable
from typing import Literal, Mapping

import pandas as pd

from lenskit.data import ID, GenericKey, ItemList, ItemListCollection, UserIDKey
from lenskit.pipeline import Pipeline, PipelineProfiler

from ._results import BatchResults
from ._runner import BatchPipelineRunner, InvocationSpec

__all__ = ["BatchPipelineRunner", "BatchResults", "InvocationSpec", "predict", "recommend"]


def predict(
    pipeline: Pipeline,
    test: ItemListCollection[GenericKey] | Mapping[ID, ItemList] | pd.DataFrame,
    *,
    n_jobs: int | Literal["ray"] | None = None,
) -> ItemListCollection[GenericKey]:
    """
    Convenience function to batch-generate rating predictions (or other per-item
    scores) from a pipeline.  This is a batch version of :func:`lenskit.predict`.

    .. note::

        If ``test`` is just a sequence of IDs, this method will still work, but
        it will score _all candidate items_ for each of the IDs.

    Stability:
        Caller
    """

    runner = BatchPipelineRunner(n_jobs=n_jobs)
    runner.predict()
    outs = runner.run(pipeline, test)
    return outs.output("predictions")  # type: ignore


def score(
    pipeline: Pipeline,
    test: ItemListCollection[GenericKey] | Mapping[ID, ItemList] | pd.DataFrame,
    *,
    n_jobs: int | Literal["ray"] | None = None,
) -> ItemListCollection[GenericKey]:
    """
    Convenience function to batch-generate personalized scores from a pipeline.
    This is a batch version of :func:`lenskit.predict`.

    .. note::

        If ``test`` is just a sequence of IDs, this method will still work, but
        it will score _all candidate items_ for each of the IDs.

    Stability:
        Caller
    """

    runner = BatchPipelineRunner(n_jobs=n_jobs)
    runner.score()
    outs = runner.run(pipeline, test)
    return outs.output("scores")  # type: ignore


def recommend(
    pipeline: Pipeline,
    users: ItemListCollection[GenericKey]
    | Mapping[ID, ItemList]
    | Iterable[ID | GenericKey]
    | pd.DataFrame,
    n: int | None = None,
    *,
    n_jobs: int | Literal["ray"] | None = None,
    profiler: PipelineProfiler | None = None,
) -> ItemListCollection[UserIDKey]:
    """
    Convenience function to batch-generate recommendations from a pipeline. This
    is a batch version of :func:`lenskit.recommend`.

    .. todo::

        Support more inputs than just user IDs.

    Stability:
        Caller
    """

    runner = BatchPipelineRunner(n_jobs=n_jobs, profiler=profiler)
    runner.recommend(n=n)
    outs = runner.run(pipeline, users)
    return outs.output("recommendations")  # type: ignore



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import logging

import numpy as np
import pandas as pd

from pytest import approx, mark, raises

from lenskit.data import ItemList, from_interactions_df
from lenskit.metrics import RunAnalysis, call_metric
from lenskit.metrics.predict import MAE, RMSE

_log = logging.getLogger(__name__)


def test_rmse_one():
    rmse = call_metric(RMSE, ItemList(["a"], scores=[1]), ItemList(["a"], rating=[1]))
    assert isinstance(rmse, float)
    assert rmse == approx(0)

    rmse = call_metric(RMSE, ItemList(["a"], scores=[1]), ItemList(["a"], rating=[2]))
    assert rmse == approx(1)

    rmse = call_metric(RMSE, ItemList(["a"], scores=[1]), ItemList(["a"], rating=[0.5]))
    assert rmse == approx(0.5)


def test_rmse_two():
    rmse = call_metric(
        RMSE, ItemList(["a", "b"], scores=[1, 2]), ItemList(["a", "b"], rating=[1, 2])
    )
    assert isinstance(rmse, float)
    assert rmse == approx(0)

    rmse = call_metric(
        RMSE, ItemList(["a", "b"], scores=[1, 1]), ItemList(["a", "b"], rating=[2, 2])
    )
    assert rmse == approx(1)

    rmse = call_metric(
        RMSE, ItemList(["a", "b"], scores=[1, 3]), ItemList(["a", "b"], rating=[3, 1])
    )
    assert rmse == approx(2)


def test_rmse_series_subset_items():
    rmse = call_metric(
        RMSE,
        ItemList(scores=[1, 3], item_ids=["a", "c"]),
        ItemList(rating=[3, 4, 1], item_ids=["a", "b", "c"]),
        missing_scores="ignore",
    )
    assert rmse == approx(2)


def test_rmse_series_missing_value_error():
    with raises(ValueError):
        call_metric(
            RMSE,
            ItemList(scores=[1, 3], item_ids=["a", "d"]),
            ItemList(rating=[3, 4, 1], item_ids=["a", "b", "c"]),
        )


def test_rmse_series_missing_value_ignore():
    rmse = call_metric(
        RMSE,
        ItemList(scores=[1, 3], item_ids=["a", "d"]),
        ItemList(rating=[3, 4, 1], item_ids=["a", "b", "c"]),
        missing_scores="ignore",
        missing_truth="ignore",
    )
    assert rmse == approx(2)


def test_mae_two():
    mae = call_metric(MAE, ItemList(["a", "b"], scores=[1, 2]), ItemList(["a", "b"], rating=[1, 2]))
    assert isinstance(mae, float)
    assert mae == approx(0)

    mae = call_metric(MAE, ItemList(["a", "b"], scores=[1, 1]), ItemList(["a", "b"], rating=[2, 2]))
    assert mae == approx(1)

    mae = call_metric(MAE, ItemList(["a", "b"], scores=[1, 3]), ItemList(["a", "b"], rating=[3, 1]))
    assert mae == approx(2)

    mae = call_metric(MAE, ItemList(["a", "b"], scores=[1, 3]), ItemList(["a", "b"], rating=[3, 2]))
    assert mae == approx(1.5)


@mark.slow
@mark.eval
def test_batch_rmse(ml_100k):
    from lenskit.basic import BiasScorer
    from lenskit.batch import predict
    from lenskit.pipeline import topn_pipeline
    from lenskit.splitting import SampleN, sample_users

    ds = from_interactions_df(ml_100k)

    bias = BiasScorer(damping=5)
    pipe = topn_pipeline(bias, predicts_ratings=True)

    split = sample_users(ds, 200, SampleN(5))
    pipe.train(split.train)

    preds = predict(pipe, split.test, n_jobs=1)

    pa = RunAnalysis()
    pa.add_metric(RMSE())
    pa.add_metric(MAE())

    metrics = pa.measure(preds, split.test)

    umdf = metrics.list_metrics(fill_missing=False)
    mdf = metrics.list_summary()

    # we should have all users
    assert len(umdf) == len(split.test)

    # we should only have users who are in the test data
    missing = set(umdf.index.tolist()) - set(k.user_id for k in split.test.keys())
    assert len(missing) == 0

    # we should not have any missing values
    assert all(umdf["RMSE"].notna())
    assert all(umdf["MAE"].notna())

    gs = metrics.global_metrics()
    _log.info("list metrics:\n%s", mdf)
    _log.info("global metrics:\n%s", gs)

    # we should have a reasonable mean
    assert umdf["RMSE"].mean() == approx(0.93, abs=0.05)
    assert mdf.loc["RMSE", "mean"] == approx(0.93, abs=0.05)

    assert umdf["MAE"].mean() == approx(0.76, abs=0.05)
    assert mdf.loc["MAE", "mean"] == approx(0.76, abs=0.05)

    # we should have global metrics
    assert gs["RMSE"] == approx(0.93, abs=0.05)
    assert gs["MAE"] == approx(0.76, abs=0.05)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Splitting data for train-test evaluation.
"""

import numpy as np

from lenskit.data import Dataset

from .holdout import HoldoutMethod, LastFrac, LastN, SampleFrac, SampleN
from .records import crossfold_records, sample_records
from .split import TTSplit
from .temporal import split_global_time, split_temporal_fraction
from .users import crossfold_users, sample_users

__all__ = [
    "TTSplit",
    "HoldoutMethod",
    "LastFrac",
    "LastN",
    "SampleFrac",
    "SampleN",
    "split_global_time",
    "split_temporal_fraction",
    "crossfold_records",
    "sample_records",
    "crossfold_users",
    "sample_users",
    "simple_test_pair",
]


def simple_test_pair(
    ratings: Dataset, n_users=200, n_rates=5, f_rates=None, rng: np.random.Generator | None = None
) -> TTSplit:
    """
    Return a single, basic train-test pair for some ratings.  This is only intended
    for convenience use in test and demos - do not use for research.
    """

    if f_rates:
        samp = SampleFrac(f_rates, rng=rng)
    else:
        samp = SampleN(n_rates, rng=rng)

    return sample_users(ratings, n_users, samp, rng=rng)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from ._monitor import Monitor, get_monitor
from ._records import RecordSink
from ._worker import WorkerContext, WorkerLogConfig, send_task

__all__ = [
    "Monitor",
    "get_monitor",
    "WorkerContext",
    "WorkerLogConfig",
    "send_task",
    "RecordSink",
]
